#!/usr/bin/env python3
#
# Copyright (c) 2022 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Script information:
# -------------------
# dpif_nl_exec_monitor.py uses the dpif_netlink_operate__:op_flow_execute USDT
# probe to receive all DPIF_OP_EXECUTE operations that are queued for
# transmission over the netlink socket. It will do some basic decoding, and if
# requested a packet dump. Note that there are also options to obtain
# additional information for the DPIF_OP_FLOW_* operations.
#
# Here is an example:
#
#   # ./dpif_op_nl_monitor.py --packet-decode decode --trace-exec-op
#   Display DPIF_OP_EXECUTE operations being queued for transmission...
#   TIME               CPU  COMM             PID        NL_SIZE
#   Display DPIF operations being queued for transmission onto the netlink...
#   TIME               CPU  COMM             PID        NL_SIZE    DPIF_...
#   2162030.641541513  5    handler4         2951007    164        DPIF_...
#      nlmsghdr  : len = 0, type = 36, flags = 1, seq = 0, pid = 0
#      genlmsghdr: cmd = DPIF_OP_EXECUTE, version = 1, reserved = 0
#      ovs_header: dp_ifindex = 21
#        > Decode OVS_PACKET_ATTR_* TLVs:
#        nla_len 46, nla_type OVS_PACKET_ATTR_PACKET[1], data: 00 00 00...
#        nla_len 20, nla_type OVS_PACKET_ATTR_KEY[2], data: 08 00 02 00...
#            > Decode OVS_KEY_ATTR_* TLVs:
#            nla_len 8, nla_type OVS_KEY_ATTR_PRIORITY[2], data: 00 00...
#            nla_len 8, nla_type OVS_KEY_ATTR_SKB_MARK[15], data: 00 00...
#        nla_len 88, nla_type OVS_PACKET_ATTR_ACTIONS[3], data: 4c 00 03...
#            > Decode OVS_ACTION_ATTR_* TLVs:
#            nla_len 76, nla_type OVS_ACTION_ATTR_SET[3], data: 48 00...
#                    > Decode OVS_TUNNEL_KEY_ATTR_* TLVs:
#                    nla_len 12, nla_type OVS_TUNNEL_KEY_ATTR_ID[0], data:...
#                    nla_len 20, nla_type OVS_TUNNEL_KEY_ATTR_IPV6_DST[13], ...
#                    nla_len 5, nla_type OVS_TUNNEL_KEY_ATTR_TTL[4], data: 40
#                    nla_len 4, nla_type OVS_TUNNEL_KEY_ATTR_DONT_FRAGMENT...
#                    nla_len 4, nla_type OVS_TUNNEL_KEY_ATTR_CSUM[6], data:
#                    nla_len 6, nla_type OVS_TUNNEL_KEY_ATTR_TP_DST[10],...
#                    nla_len 12, nla_type OVS_TUNNEL_KEY_ATTR_GENEVE_OPTS...
#            nla_len 8, nla_type OVS_ACTION_ATTR_OUTPUT[1], data: 02 00 00 00
#        - Dumping OVS_PACKET_ATR_PACKET data:
#        ###[ Ethernet ]###
#          dst       = 00:00:00:00:ec:01
#          src       = 04:f4:bc:28:57:00
#          type      = IPv4
#        ###[ IP ]###
#             version   = 4
#             ihl       = 5
#             tos       = 0x0
#             len       = 50
#             id        = 0
#             flags     =
#             frag      = 0
#             ttl       = 127
#             proto     = icmp
#             chksum    = 0x2767
#             src       = 10.0.0.1
#             dst       = 10.0.0.100
#             \options   \
#        ###[ ICMP ]###
#                type      = echo-request
#                code      = 0
#                chksum    = 0xf7f3
#                id        = 0x0
#                seq       = 0xc
#
# The example above dumps the full netlink and packet decode. However options
# exist to disable this. For a complete list of options, please use the
# '--help' or '-h' argument.
#

from bcc import BPF, USDT, USDTException
from os.path import exists
from scapy.all import hexdump, wrpcap
from scapy.layers.l2 import Ether

import argparse
import psutil
import re
import struct
import sys

#
# Actual eBPF source code
#
ebpf_source = """
#include <linux/sched.h>

#define MAX_NLMSG <MAX_NLMSG_VAL>

struct event_t {
    u32 cpu;
    u32 pid;
    u64 ts;
    u32 nl_size;
    char comm[TASK_COMM_LEN];
    u8 nl_msg[MAX_NLMSG];
};

struct ofpbuf {
    void *base;
    void *data;
    uint32_t size;

    /* The actual structure is longer, but we are only interested in the
     * first couple of entries. */
};

BPF_RINGBUF_OUTPUT(events, <BUFFER_PAGE_CNT>);
BPF_TABLE("percpu_array", uint32_t, uint64_t, dropcnt, 1);

static int trace_event(struct ofpbuf *nlbuf)
{
    uint32_t size;

    struct event_t *event = events.ringbuf_reserve(sizeof(struct event_t));
    if (!event) {
        uint32_t type = 0;
        uint64_t *value = dropcnt.lookup(&type);
        if (value)
            __sync_fetch_and_add(value, 1);

        return 1;
    }

    event->ts = bpf_ktime_get_ns();
    event->cpu =  bpf_get_smp_processor_id();
    event->pid = bpf_get_current_pid_tgid();
    bpf_get_current_comm(&event->comm, sizeof(event->comm));

    event->nl_size = nlbuf->size;
    if (event->nl_size > MAX_NLMSG)
        size = MAX_NLMSG;
    else
        size = event->nl_size;

    bpf_probe_read(&event->nl_msg, size, nlbuf->data);

    events.ringbuf_submit(event, 0);
    return 0;
}

#if <ENABLE_OP_EXECUTE>
int trace__op_execute(struct pt_regs *ctx) {
    struct ofpbuf nlbuf;

    bpf_usdt_readarg_p(5, ctx, &nlbuf, sizeof(nlbuf));
    return trace_event(&nlbuf);
};
#endif

#if <ENABLE_OP_FLOW_PUT>
int trace__op_flow_put(struct pt_regs *ctx) {
    struct ofpbuf nlbuf;

    bpf_usdt_readarg_p(4, ctx, &nlbuf, sizeof(nlbuf));
    return trace_event(&nlbuf);
};
#endif

#if <ENABLE_OP_FLOW_DEL>
int trace__op_flow_del(struct pt_regs *ctx) {
    struct ofpbuf nlbuf;

    bpf_usdt_readarg_p(4, ctx, &nlbuf, sizeof(nlbuf));
    return trace_event(&nlbuf);
};
#endif

#if <ENABLE_OP_FLOW_GET>
int trace__op_flow_get(struct pt_regs *ctx) {
    struct ofpbuf nlbuf;

    bpf_usdt_readarg_p(4, ctx, &nlbuf, sizeof(nlbuf));
    return trace_event(&nlbuf);
};
#endif
"""


#
# print_event()
#
def print_event(ctx, data, size):
    event = b["events"].event(data)

    if event.nl_size < options.nlmsg_size:
        nl_size = event.nl_size
    else:
        nl_size = options.nlmsg_size

    print("{:<18.9f} {:<4} {:<16} {:<10} {:<10} {}".
          format(event.ts / 1000000000,
                 event.cpu,
                 event.comm.decode("utf-8"),
                 event.pid,
                 event.nl_size,
                 get_ovs_dpif_op_str(get_cmd_type_from_nlm(
                     bytes(event.nl_msg)[:nl_size]))))

    #
    # Dumping the netlink message data if requested.
    #
    if options.nlmsg_decode == "hex":
        #
        # Abuse scapy's hex dump to dump flow key
        #
        print(re.sub("^", " " * 4,
                     hexdump(Ether(bytes(event.nl_msg)[:nl_size]), dump=True),
                     flags=re.MULTILINE))

    if options.nlmsg_decode == "nlraw":
        decode_result = decode_nlm(bytes(event.nl_msg)[:nl_size], dump=True)
    else:
        decode_result = decode_nlm(bytes(event.nl_msg)[:nl_size], dump=False)

    #
    # Decode packet only if there is data
    #
    if "OVS_PACKET_ATTR_PACKET" not in decode_result:
        return

    pkt_data = decode_result["OVS_PACKET_ATTR_PACKET"]
    indent = 4 if options.nlmsg_decode != "nlraw" else 6

    if options.packet_decode != "none":
        print("{}- Dumping OVS_PACKET_ATR_PACKET data:".format(" " * indent))

    if options.packet_decode == "hex":
        print(re.sub("^", " " * indent, hexdump(pkt_data, dump=True),
                     flags=re.MULTILINE))

    packet = Ether(pkt_data)
    if options.packet_decode == "decode":
        print(re.sub("^", " " * indent, packet.show(dump=True),
                     flags=re.MULTILINE))

    if options.pcap is not None:
        wrpcap(options.pcap, packet, append=True)


#
# decode_nlm_tlvs()
#
def decode_nlm_tlvs(tlvs, header=None, indent=4, dump=True,
                    attr_to_str_func=None, decode_tree=None):
    bytes_left = len(tlvs)
    result = {}

    if dump:
        print("{}{}".format(" " * indent, header))

    while bytes_left:
        if bytes_left < 4:
            if dump:
                print("{}WARN: decode truncated; can't read header".format(
                    " " * indent))
            break

        nla_len, nla_type = struct.unpack("=HH", tlvs[:4])

        if nla_len < 4:
            if dump:
                print("{}WARN: decode truncated; nla_len < 4".format(
                    " " * indent))
            break

        nla_data = tlvs[4:nla_len]
        trunc = ""

        if attr_to_str_func is None:
            nla_type_name = "type_{}".format(nla_type)
        else:
            nla_type_name = attr_to_str_func(nla_type)

        if nla_len > bytes_left:
            trunc = "..."
            nla_data = nla_data[:(bytes_left - 4)]
        else:
            result[nla_type_name] = nla_data

        if dump:
            print("{}nla_len {}, nla_type {}[{}], data: {}{}".format(
                " " * indent, nla_len, nla_type_name, nla_type,
                "".join("{:02x} ".format(b) for b in nla_data), trunc))

            #
            # If we have the full data, try to decode further
            #
            if trunc == "" and decode_tree is not None \
               and nla_type_name in decode_tree:
                node = decode_tree[nla_type_name]
                decode_nlm_tlvs(nla_data,
                                header=node["header"],
                                indent=indent + node["indent"], dump=True,
                                attr_to_str_func=node["attr_str_func"],
                                decode_tree=node["decode_tree"])

        if trunc != "":
            if dump:
                print("{}WARN: decode truncated; nla_len > msg_len[{}] ".
                      format(" " * indent, bytes_left))
            break

        # update next offset, but make sure it's aligned correctly
        next_offset = (nla_len + 3) & ~(3)
        tlvs = tlvs[next_offset:]
        bytes_left -= next_offset

    return result


#
# get_cmd_type_from_nlm()
#
def get_cmd_type_from_nlm(nlm):
    # The netlink message consists of at least a 'struct nlmsghdr' (16-bytes)
    # followed by a 'struct genlmsghdr'. The first byte of the genlmsghdr
    # structure contains the command. Which is what we will extract here.
    if len(nlm) < 17:
        return -1

    return nlm[16]


#
# decode_nlm()
#
def decode_nlm(msg, indent=4, dump=True):
    result = {}

    #
    # Decode 'struct nlmsghdr'
    #
    if dump:
        print("{}nlmsghdr  : len = {}, type = {}, flags = {}, seq = {}, "
              "pid = {}".format(" " * indent,
                                *struct.unpack("=IHHII", msg[:16])))

    msg = msg[16:]

    #
    # Decode 'struct genlmsghdr'
    #
    cmd, version, reserved = struct.unpack("=BBH", msg[:4])

    if dump:
        print("{}genlmsghdr: cmd = {}, version = {}, reserved = {}".format(
            " " * indent, get_ovs_dpif_op_str(cmd), version, reserved))

    msg = msg[4:]

    #
    # Decode 'struct ovs_header'
    #
    if dump:
        print("{}ovs_header: dp_ifindex = {}".format(
            " " * indent, *struct.unpack("=I", msg[:4])))

    msg = msg[4:]

    #
    # Decode TLVs
    #
    nl_key_attr = {
        "header": "> Decode OVS_KEY_ATTR_* TLVs:",
        "indent": 4,
        "attr_str_func": get_ovs_key_attr_str,
        "decode_tree": {
            "OVS_KEY_ATTR_ENCAP": {
                "header": "> Decode OVS_KEY_ATTR_* TLVs:",
                "indent": 4,
                "attr_str_func": get_ovs_key_attr_str,
                "decode_tree": {
                    "OVS_KEY_ATTR_ENCAP": {
                        "header": "> Decode OVS_KEY_ATTR_* TLVs:",
                        "indent": 4,
                        "attr_str_func": get_ovs_key_attr_str,
                        "decode_tree": None,
                    },
                },
            },
        }
    }

    nl_action_attr = {
        "header": "> Decode OVS_ACTION_ATTR_* TLVs:",
        "indent": 4,
        "attr_str_func": get_ovs_action_attr_str,
        "decode_tree": {
            "OVS_ACTION_ATTR_SET": {
                "header": "> Decode OVS_KEY_ATTR_* TLVs:",
                "indent": 4,
                "attr_str_func": get_ovs_key_attr_str,
                "decode_tree": {
                    "OVS_KEY_ATTR_TUNNEL": {
                        "header": "> Decode OVS_TUNNEL_KEY_ATTR_* TLVs:",
                        "indent": 4,
                        "attr_str_func": get_ovs_tunnel_key_attr_str,
                        "decode_tree": None,
                    },
                },
            },
        },
    }

    nl_attr_tree_exec = {
        "OVS_PACKET_ATTR_KEY": nl_key_attr,
        "OVS_PACKET_ATTR_ACTIONS": nl_action_attr,
    }

    nl_attr_tree_put = {
        "OVS_FLOW_ATTR_KEY": nl_key_attr,
        "OVS_FLOW_ATTR_MASK": nl_key_attr,
        "OVS_FLOW_ATTR_ACTIONS": nl_action_attr,
    }

    if get_ovs_dpif_op_str(cmd) == "DPIF_OP_EXECUTE":
        result = decode_nlm_tlvs(msg, indent=indent + 2, dump=dump,
                                 header="> Decode OVS_PACKET_ATTR_* TLVs:",
                                 attr_to_str_func=get_ovs_pkt_attr_str,
                                 decode_tree=nl_attr_tree_exec)
    else:
        result = decode_nlm_tlvs(msg, indent=indent + 2, dump=dump,
                                 header="> Decode OVS_FLOW_ATTR_* TLVs:",
                                 attr_to_str_func=get_ovs_flow_attr_str,
                                 decode_tree=nl_attr_tree_put)
    return result


#
# get_ovs_flow_attr_str()
#
def get_ovs_flow_attr_str(attr):
    ovs_flow_attr = ["OVS_FLOW_ATTR_UNSPEC",
                     "OVS_FLOW_ATTR_KEY",
                     "OVS_FLOW_ATTR_ACTIONS",
                     "OVS_FLOW_ATTR_STATS",
                     "OVS_FLOW_ATTR_TCP_FLAGS",
                     "OVS_FLOW_ATTR_USED",
                     "OVS_FLOW_ATTR_CLEAR",
                     "OVS_FLOW_ATTR_MASK",
                     "OVS_FLOW_ATTR_PROBE",
                     "OVS_FLOW_ATTR_UFID",
                     "OVS_FLOW_ATTR_UFID_FLAGS",
                     "OVS_FLOW_ATTR_PAD"]
    if attr < 0 or attr >= len(ovs_flow_attr):
        return "<UNKNOWN:{}>".format(attr)

    return ovs_flow_attr[attr]


#
# get_ovs_pkt_attr_str()
#
def get_ovs_pkt_attr_str(attr):
    ovs_pkt_attr = ["OVS_PACKET_ATTR_UNSPEC",
                    "OVS_PACKET_ATTR_PACKET",
                    "OVS_PACKET_ATTR_KEY",
                    "OVS_PACKET_ATTR_ACTIONS",
                    "OVS_PACKET_ATTR_USERDATA",
                    "OVS_PACKET_ATTR_EGRESS_TUN_KEY",
                    "OVS_PACKET_ATTR_UNUSED1",
                    "OVS_PACKET_ATTR_UNUSED2",
                    "OVS_PACKET_ATTR_PROBE",
                    "OVS_PACKET_ATTR_MRU",
                    "OVS_PACKET_ATTR_LEN",
                    "OVS_PACKET_ATTR_HASH"]
    if attr < 0 or attr >= len(ovs_pkt_attr):
        return "<UNKNOWN:{}>".format(attr)

    return ovs_pkt_attr[attr]


#
# get_ovs_key_attr_str()
#
def get_ovs_key_attr_str(attr):
    ovs_key_attr = ["OVS_KEY_ATTR_UNSPEC",
                    "OVS_KEY_ATTR_ENCAP",
                    "OVS_KEY_ATTR_PRIORITY",
                    "OVS_KEY_ATTR_IN_PORT",
                    "OVS_KEY_ATTR_ETHERNET",
                    "OVS_KEY_ATTR_VLAN",
                    "OVS_KEY_ATTR_ETHERTYPE",
                    "OVS_KEY_ATTR_IPV4",
                    "OVS_KEY_ATTR_IPV6",
                    "OVS_KEY_ATTR_TCP",
                    "OVS_KEY_ATTR_UDP",
                    "OVS_KEY_ATTR_ICMP",
                    "OVS_KEY_ATTR_ICMPV6",
                    "OVS_KEY_ATTR_ARP",
                    "OVS_KEY_ATTR_ND",
                    "OVS_KEY_ATTR_SKB_MARK",
                    "OVS_KEY_ATTR_TUNNEL",
                    "OVS_KEY_ATTR_SCTP",
                    "OVS_KEY_ATTR_TCP_FLAGS",
                    "OVS_KEY_ATTR_DP_HASH",
                    "OVS_KEY_ATTR_RECIRC_ID",
                    "OVS_KEY_ATTR_MPLS",
                    "OVS_KEY_ATTR_CT_STATE",
                    "OVS_KEY_ATTR_CT_ZONE",
                    "OVS_KEY_ATTR_CT_MARK",
                    "OVS_KEY_ATTR_CT_LABELS",
                    "OVS_KEY_ATTR_CT_ORIG_TUPLE_IPV4",
                    "OVS_KEY_ATTR_CT_ORIG_TUPLE_IPV6",
                    "OVS_KEY_ATTR_NSH"]

    if attr < 0 or attr >= len(ovs_key_attr):
        return "<UNKNOWN:{}>".format(attr)

    return ovs_key_attr[attr]


#
# get_ovs_action_attr_str()
#
def get_ovs_action_attr_str(attr):
    ovs_action_attr = ["OVS_ACTION_ATTR_UNSPEC",
                       "OVS_ACTION_ATTR_OUTPUT",
                       "OVS_ACTION_ATTR_USERSPACE",
                       "OVS_ACTION_ATTR_SET",
                       "OVS_ACTION_ATTR_PUSH_VLAN",
                       "OVS_ACTION_ATTR_POP_VLAN",
                       "OVS_ACTION_ATTR_SAMPLE",
                       "OVS_ACTION_ATTR_RECIRC",
                       "OVS_ACTION_ATTR_HASH",
                       "OVS_ACTION_ATTR_PUSH_MPLS",
                       "OVS_ACTION_ATTR_POP_MPLS",
                       "OVS_ACTION_ATTR_SET_MASKED",
                       "OVS_ACTION_ATTR_CT",
                       "OVS_ACTION_ATTR_TRUNC",
                       "OVS_ACTION_ATTR_PUSH_ETH",
                       "OVS_ACTION_ATTR_POP_ETH",
                       "OVS_ACTION_ATTR_CT_CLEAR",
                       "OVS_ACTION_ATTR_PUSH_NSH",
                       "OVS_ACTION_ATTR_POP_NSH",
                       "OVS_ACTION_ATTR_METER",
                       "OVS_ACTION_ATTR_CLONE",
                       "OVS_ACTION_ATTR_CHECK_PKT_LEN",
                       "OVS_ACTION_ATTR_ADD_MPLS",
                       "OVS_ACTION_ATTR_DEC_TTL",
                       "OVS_ACTION_ATTR_DROP",
                       "OVS_ACTION_ATTR_PSAMPLE",
                       "OVS_ACTION_ATTR_TUNNEL_PUSH",
                       "OVS_ACTION_ATTR_TUNNEL_POP",
                       "OVS_ACTION_ATTR_LB_OUTPUT"]
    if attr < 0 or attr >= len(ovs_action_attr):
        return "<UNKNOWN:{}>".format(attr)

    return ovs_action_attr[attr]


#
# get_ovs_tunnel_key_attr_str()
#
def get_ovs_tunnel_key_attr_str(attr):
    ovs_tunnel_key_attr = ["OVS_TUNNEL_KEY_ATTR_ID",
                           "OVS_TUNNEL_KEY_ATTR_IPV4_SRC",
                           "OVS_TUNNEL_KEY_ATTR_IPV4_DST",
                           "OVS_TUNNEL_KEY_ATTR_TOS",
                           "OVS_TUNNEL_KEY_ATTR_TTL",
                           "OVS_TUNNEL_KEY_ATTR_DONT_FRAGMENT",
                           "OVS_TUNNEL_KEY_ATTR_CSUM",
                           "OVS_TUNNEL_KEY_ATTR_OAM",
                           "OVS_TUNNEL_KEY_ATTR_GENEVE_OPTS",
                           "OVS_TUNNEL_KEY_ATTR_TP_SRC",
                           "OVS_TUNNEL_KEY_ATTR_TP_DST",
                           "OVS_TUNNEL_KEY_ATTR_VXLAN_OPTS",
                           "OVS_TUNNEL_KEY_ATTR_IPV6_SRC",
                           "OVS_TUNNEL_KEY_ATTR_IPV6_DST",
                           "OVS_TUNNEL_KEY_ATTR_PAD",
                           "OVS_TUNNEL_KEY_ATTR_ERSPAN_OPTS",
                           "OVS_TUNNEL_KEY_ATTR_GTPU_OPTS"]
    if attr < 0 or attr >= len(ovs_tunnel_key_attr):
        return "<UNKNOWN:{}>".format(attr)

    return ovs_tunnel_key_attr[attr]


#
# get_ovs_dpif_op_str()
#
def get_ovs_dpif_op_str(op):
    ovs_dpif_ops = ["DPIF_OP_UNSPEC",
                    "DPIF_OP_FLOW_PUT",
                    "DPIF_OP_FLOW_DEL",
                    "DPIF_OP_EXECUTE",
                    "DPIF_OP_FLOW_GET"]

    if op < 0 or op >= len(ovs_dpif_ops):
        return "<UNKNOWN:{}>".format(op)

    return ovs_dpif_ops[op]


#
# buffer_size_type()
#
def buffer_size_type(astr, min=64, max=2048):
    value = int(astr)
    if min <= value <= max:
        return value
    else:
        raise argparse.ArgumentTypeError(
            "value not in range {}-{}".format(min, max))


#
# next_power_of_two()
#
def next_power_of_two(val):
    np = 1
    while np < val:
        np *= 2
    return np


#
# main()
#
def main():
    #
    # Don't like these globals, but ctx passing does not seem to work with the
    # existing open_ring_buffer() API :(
    #
    global b
    global options

    #
    # Argument parsing
    #
    parser = argparse.ArgumentParser()

    parser.add_argument("--buffer-page-count",
                        help="Number of BPF ring buffer pages, default 1024",
                        type=int, default=1024, metavar="NUMBER")
    parser.add_argument("-D", "--debug",
                        help="Enable eBPF debugging",
                        type=int, const=0x3f, default=0, nargs="?")
    parser.add_argument("-d", "--packet-decode",
                        help="Display packet content in selected mode, "
                        "default none",
                        choices=["none", "hex", "decode"], default="none")
    parser.add_argument("-n", "--nlmsg-decode",
                        help="Display netlink message content in selected mode"
                        ", default nlraw",
                        choices=["none", "hex", "nlraw"], default="nlraw")
    parser.add_argument("-p", "--pid", metavar="VSWITCHD_PID",
                        help="ovs-vswitch's PID",
                        type=int, default=None)
    parser.add_argument("-s", "--nlmsg-size",
                        help="Set maximum netlink message size to capture, "
                        "default 512", type=buffer_size_type, default=512,
                        metavar="[64-2048]")
    parser.add_argument("--trace-del-op",
                        help="Monitor DPIF_OP_FLOW_DEL messages",
                        action="store_true")
    parser.add_argument("--trace-exec-op",
                        help="Monitor DPIF_OP_EXECUTE messages",
                        action="store_true")
    parser.add_argument("--trace-get-op",
                        help="Monitor DPIF_OP_FLOW_GET messages",
                        action="store_true")
    parser.add_argument("--trace-put-op",
                        help="Monitor DPIF_OP_FLOW_PUT messages",
                        action="store_true")
    parser.add_argument("-w", "--pcap", metavar="PCAP_FILE",
                        help="Write execute packets to specified pcap file",
                        type=str, default=None)

    options = parser.parse_args()

    #
    # Verify arguments.
    #
    if (not options.trace_del_op and not options.trace_exec_op
       and not options.trace_get_op and not options.trace_put_op):
        print("ERROR: At least on of the '--trace-*-op' arguments should be "
              "specified!")
        sys.exit(-1)

    #
    # Find the PID of the ovs-vswitchd daemon if not specified.
    #
    if options.pid is None:
        for proc in psutil.process_iter():
            if "ovs-vswitchd" in proc.name():
                if options.pid is not None:
                    print("ERROR: Multiple ovs-vswitchd daemons running, "
                          "use the -p option!")
                    sys.exit(-1)

                options.pid = proc.pid

    #
    # Error checking on input parameters
    #
    if options.pid is None:
        print("ERROR: Failed to find ovs-vswitchd's PID!")
        sys.exit(-1)

    if options.pcap is not None:
        if exists(options.pcap):
            print("ERROR: Destination capture file \"{}\" already exists!".
                  format(options.pcap))
            sys.exit(-1)

    options.buffer_page_count = next_power_of_two(options.buffer_page_count)

    #
    # Attach the usdt probe
    #
    u = USDT(pid=int(options.pid))
    try:
        if options.trace_exec_op:
            u.enable_probe(probe="dpif_netlink_operate__:op_flow_execute",
                           fn_name="trace__op_execute")
        if options.trace_put_op:
            u.enable_probe(probe="dpif_netlink_operate__:op_flow_put",
                           fn_name="trace__op_flow_put")
        if options.trace_del_op:
            u.enable_probe(probe="dpif_netlink_operate__:op_flow_del",
                           fn_name="trace__op_flow_del")
        if options.trace_get_op:
            u.enable_probe(probe="dpif_netlink_operate__:op_flow_get",
                           fn_name="trace__op_flow_get")
    except USDTException as e:
        print("ERROR: {}".format(
            (re.sub("^", " " * 7, str(e), flags=re.MULTILINE)).strip().
            replace("--with-dtrace or --enable-dtrace",
                    "--enable-usdt-probes")))
        sys.exit(-1)

    #
    # Uncomment to see how arguments are decoded.
    #   print(u.get_text())
    #

    #
    # Attach probe to running process
    #
    source = ebpf_source.replace("<MAX_NLMSG_VAL>", str(options.nlmsg_size))
    source = source.replace("<BUFFER_PAGE_CNT>",
                            str(options.buffer_page_count))

    source = source.replace("<ENABLE_OP_FLOW_DEL>",
                            "1" if options.trace_del_op else "0")
    source = source.replace("<ENABLE_OP_EXECUTE>",
                            "1" if options.trace_exec_op else "0")
    source = source.replace("<ENABLE_OP_FLOW_GET>",
                            "1" if options.trace_get_op else "0")
    source = source.replace("<ENABLE_OP_FLOW_PUT>",
                            "1" if options.trace_put_op else "0")

    b = BPF(text=source, usdt_contexts=[u], debug=options.debug)

    #
    # Print header
    #
    print("Display DPIF operations being queued for transmission onto the "
          "netlink socket.")
    print("{:<18} {:<4} {:<16} {:<10} {:<10} {}".format(
        "TIME", "CPU", "COMM", "PID", "NL_SIZE", "DPIF_OPERATION"))

    #
    # Dump out all events
    #
    b["events"].open_ring_buffer(print_event)
    while 1:
        try:
            b.ring_buffer_poll()
        except KeyboardInterrupt:
            break

    dropcnt = b.get_table("dropcnt")
    for k in dropcnt.keys():
        count = dropcnt.sum(k).value
        if k.value == 0 and count > 0:
            print("\nWARNING: Not all upcalls were captured, {} were dropped!"
                  "\n         Increase the BPF ring buffer size with the "
                  "--buffer-page-count option.".format(count))


#
# Start main() as the default entry point...
#
if __name__ == "__main__":
    main()



# Copyright (c) 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import click

from ovs.flowviz.main import maincli
from ovs.flowviz.ofp.html import HTMLProcessor
from ovs.flowviz.ofp.logic import CookieProcessor, LogicFlowProcessor
from ovs.flowviz.process import (
    ConsoleProcessor,
    JSONOpenFlowProcessor,
)


@maincli.group(subcommand_metavar="FORMAT")
@click.pass_obj
def openflow(opts):
    """Process OpenFlow Flows."""
    pass


@openflow.command()
@click.pass_obj
def json(opts):
    """Print the flows in JSON format."""
    proc = JSONOpenFlowProcessor(opts)
    proc.process()
    print(proc.json_string())


@openflow.command()
@click.option(
    "-h",
    "--heat-map",
    is_flag=True,
    default=False,
    show_default=True,
    help="Create heat-map with packet and byte counters",
)
@click.pass_obj
def console(opts, heat_map):
    """Print the flows in the console with some style."""
    proc = ConsoleProcessor(
        opts,
        "ofp",
        heat_map=["n_packets", "n_bytes"] if heat_map else [],
    )
    proc.process()
    proc.print()


def ovn_detrace_callback(ctx, param, value):
    """click callback to add detrace information to config object and
    set general ovn-detrace flag to True
    """
    ctx.obj[param.name] = value
    if value != param.default:
        ctx.obj["ovn_detrace_flag"] = True
    return value


@openflow.command()
@click.option(
    "-d",
    "--ovn-detrace",
    "ovn_detrace_flag",
    is_flag=True,
    show_default=True,
    help="Use ovn-detrace to extract cookie information (implies '-c')",
)
@click.option(
    "--ovn-detrace-path",
    default="/usr/bin",
    type=click.Path(),
    help="Use an alternative path to where ovn_detrace.py is located. "
    "Instead of using this option you can just set PYTHONPATH accordingly.",
    show_default=True,
    callback=ovn_detrace_callback,
)
@click.option(
    "--ovnnb-db",
    default=os.getenv("OVN_NB_DB") or "unix:/var/run/ovn/ovnnb_db.sock",
    help="Specify the OVN NB database string (implies -d). "
    "If the OVN_NB_DB environment variable is set, it's used as default. "
    "Otherwise, the default is unix:/var/run/ovn/ovnnb_db.sock",
    callback=ovn_detrace_callback,
)
@click.option(
    "--ovnsb-db",
    default=os.getenv("OVN_SB_DB") or "unix:/var/run/ovn/ovnsb_db.sock",
    help="Specify the OVN NB database string (implies -d). "
    "If the OVN_NB_DB environment variable is set, it's used as default. "
    "Otherwise, the default is unix:/var/run/ovn/ovnnb_db.sock",
    callback=ovn_detrace_callback,
)
@click.option(
    "-o",
    "--ovn-filter",
    help="Specify a filter to be run on ovn-detrace information (implied -d). "
    "Format: python regular expression "
    "(see https://docs.python.org/3/library/re.html)",
    callback=ovn_detrace_callback,
)
@click.option(
    "-s",
    "--show-flows",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show the full flows under each logical flow",
)
@click.option(
    "-c",
    "--cookie",
    "cookie_flag",
    is_flag=True,
    default=False,
    show_default=True,
    help="Consider the cookie in the logical flow",
)
@click.option(
    "-h",
    "--heat-map",
    is_flag=True,
    default=False,
    show_default=True,
    help="Create heat-map with packet and byte counters (when -s is used)",
)
@click.pass_obj
def logic(
    opts,
    ovn_detrace_flag,
    ovn_detrace_path,
    ovnnb_db,
    ovnsb_db,
    ovn_filter,
    show_flows,
    cookie_flag,
    heat_map,
):
    """
    Print the logical structure of the flows.

    First, sorts the flows based on tables and priorities.
    Then, deduplicates logically equivalent flows: these a flows that match
    on the same set of fields (regardless of the values they match against),
    have the same priority, and actions (regardless of action arguments,
    except in the case of output and recirculate).
    Optionally, the cookie can also be considered to be part of the logical
    flow.
    """
    if ovn_detrace_flag:
        opts["ovn_detrace_flag"] = True
    if opts.get("ovn_detrace_flag"):
        cookie_flag = True

    processor = LogicFlowProcessor(opts, cookie_flag, heat_map)
    processor.process()
    processor.print(show_flows)


@openflow.command()
@click.option(
    "-d",
    "--ovn-detrace",
    "ovn_detrace_flag",
    is_flag=True,
    show_default=True,
    help="Use ovn-detrace to extract cookie information",
)
@click.option(
    "--ovn-detrace-path",
    default="/usr/bin",
    type=click.Path(),
    help="Use an alternative path to where ovn_detrace.py is located. "
    "Instead of using this option you can just set PYTHONPATH accordingly",
    show_default=True,
    callback=ovn_detrace_callback,
)
@click.option(
    "--ovnnb-db",
    default=os.getenv("OVN_NB_DB") or "unix:/var/run/ovn/ovnnb_db.sock",
    help="Specify the OVN NB database string (implies -d). "
    "If the OVN_NB_DB environment variable is set, it's used as default. "
    "Otherwise, the default is unix:/var/run/ovn/ovnnb_db.sock",
    callback=ovn_detrace_callback,
)
@click.option(
    "--ovnsb-db",
    default=os.getenv("OVN_SB_DB") or "unix:/var/run/ovn/ovnsb_db.sock",
    help="Specify the OVN NB database string (implies -d). "
    "If the OVN_NB_DB environment variable is set, it's used as default. "
    "Otherwise, the default is unix:/var/run/ovn/ovnnb_db.sock",
    callback=ovn_detrace_callback,
)
@click.option(
    "-o",
    "--ovn-filter",
    help="Specify a filter to be run on ovn-detrace information (implied -d). "
    "Format: python regular expression "
    "(see https://docs.python.org/3/library/re.html)",
    callback=ovn_detrace_callback,
)
@click.pass_obj
def cookie(
    opts, ovn_detrace_flag, ovn_detrace_path, ovnnb_db, ovnsb_db, ovn_filter
):
    """Print the flow tables sorted by cookie."""
    if ovn_detrace_flag:
        opts["ovn_detrace_flag"] = True

    processor = CookieProcessor(opts)
    processor.process()
    processor.print()


@openflow.command()
@click.pass_obj
def html(opts):
    """Print the flows in an linked HTML list arranged by tables."""
    processor = HTMLProcessor(opts)
    processor.process()
    print(processor.html())



#!/usr/bin/env python3
#
# Copyright (c) 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Script information:
# -------------------
# upcall_monitor.py uses the dpif_recv:recv_upcall USDT to receive all upcall
# packets sent by the kernel to ovs-vswitchd. By default, it will show all
# upcall events, which looks something like this:
#
# TIME               CPU  COMM      PID      PORT_NAME                TYPE ..
# 5952147.003848809  2    handler4  1381158  eth0 (system@ovs-system)  0
# 5952147.003879643  2    handler4  1381158  eth0 (system@ovs-system)  0
# 5952147.003914924  2    handler4  1381158  eth0 (system@ovs-system)  0
#
# Also, upcalls dropped by the kernel (e.g: because the netlink buffer is full)
# are reported. This requires the kernel version to be greater or equal to
# 5.14.
# In addition, the packet and flow key data can be dumped. This can be done
# using the --packet-decode and --flow-key decode options (see below).
#
# Note that by default only 64 bytes of the packet and flow key are retrieved.
# If you would like to capture all or more of the packet and/or flow key data,
# the ----packet-size and --flow-key-size options can be used.
#
# If required, the received packets can also be stored in a pcap file using the
# --pcap option.
#
# The following are the available options:
#
#    usage: upcall_monitor.py [-h] [-D [DEBUG]] [-d {none,hex,decode}]
#                             [-f [64-2048]] [-k {none,hex,nlraw}]
#                             [-p VSWITCHD_PID] [-s [64-2048]] [-w PCAP_FILE]
#
#    optional arguments:
#      -h, --help            show this help message and exit
#      -D [DEBUG], --debug [DEBUG]
#                            Enable eBPF debugging
#      -d {none,hex,decode}, --packet-decode {none,hex,decode}
#                            Display packet content in selected mode,
#                            default none
#      -f [64-2048], --flow-key-size [64-2048]
#                            Set maximum flow key size to capture, default 64
#      -k {none,hex,nlraw}, --flow-key-decode {none,hex,nlraw}
#                            Display flow-key content in selected mode, default
#                            none
#      -p VSWITCHD_PID, --pid VSWITCHD_PID
#                            ovs-vswitch's PID
#      -s [64-2048], --packet-size [64-2048]
#                            Set maximum packet size to capture, default 64
#      -w PCAP_FILE, --pcap PCAP_FILE
#                            Write upcall packets to specified pcap file.
#      -r, --result {error,ok,any}
#                            Display only events with the given result,
#                            default: any
#
# The following is an example of how to use the script on the running
# ovs-vswitchd process with a packet and flow key dump enabled:
#
#  $ ./upcall_monitor.py --packet-decode decode --flow-key-decode nlraw \
#      --packet-size 128 --flow-key-size 256
#  TIME               CPU  COMM             PID        PORT_NAME          ...
#  5953013.333214231  2    handler4         1381158    system@ovs-system  ...
#    Flow key size 132 bytes, size captured 132 bytes.
#      nla_len 8, nla_type OVS_KEY_ATTR_RECIRC_ID[20], data: 00 00 00 00
#      nla_len 8, nla_type OVS_KEY_ATTR_DP_HASH[19], data: 00 00 00 00
#      nla_len 8, nla_type OVS_KEY_ATTR_PRIORITY[2], data: 00 00 00 00
#      nla_len 8, nla_type OVS_KEY_ATTR_IN_PORT[3], data: 02 00 00 00
#      nla_len 8, nla_type OVS_KEY_ATTR_SKB_MARK[15], data: 00 00 00 00
#      nla_len 8, nla_type OVS_KEY_ATTR_CT_STATE[22], data: 00 00 00 00
#      nla_len 6, nla_type OVS_KEY_ATTR_CT_ZONE[23], data: 00 00
#      nla_len 8, nla_type OVS_KEY_ATTR_CT_MARK[24], data: 00 00 00 00
#      nla_len 20, nla_type OVS_KEY_ATTR_CT_LABELS[25], data: 00 00 00 00 ...
#      nla_len 16, nla_type OVS_KEY_ATTR_ETHERNET[4], data: 04 f4 bc 28 57 ...
#      nla_len 6, nla_type OVS_KEY_ATTR_ETHERTYPE[6], data: 08 00
#      nla_len 16, nla_type OVS_KEY_ATTR_IPV4[7], data: 01 01 01 64 01 01 ...
#      nla_len 6, nla_type OVS_KEY_ATTR_ICMP[11], data: 00 00
#    1: Receive dp_port 2, packet size 98 bytes, size captured 98 bytes.
#      ###[ Ethernet ]###
#        dst       = 3c:fd:fe:9e:7f:68
#        src       = 04:f4:bc:28:57:01
#        type      = IPv4
#      ###[ IP ]###
#           version   = 4
#           ihl       = 5
#           tos       = 0x0
#           len       = 84
#           id        = 41404
#           flags     = DF
#           frag      = 0
#           ttl       = 64
#           proto     = icmp
#           chksum    = 0x940c
#           src       = 1.1.1.100
#           dst       = 1.1.1.123
#           \options   \
#      ###[ ICMP ]###
#              type      = echo-reply
#              code      = 0
#              chksum    = 0x2f55
#              id        = 0x90e6
#              seq       = 0x1
#      ###[ Raw ]###
#                 load      = 'GBTa\x00\x00\x00\x00\xd8L\r\x00\x00\x00\x00\...
#

from bcc import BPF, USDT, USDTException
from os.path import exists, join
from scapy import VERSION as scapy_version
from scapy.layers.l2 import Ether
from usdt_lib import DpPortMapping

import argparse
import psutil
import re
import struct
import sys

(scapy_mayor, scapy_minor, _) = scapy_version.split(".", 2)
(scapy_mayor, scapy_minor) = (int(scapy_mayor), int(scapy_minor))

scapy_supports_pcap_iface = False
if scapy_mayor < 2 or (scapy_mayor == 2 and scapy_minor) <= 4:
    from scapy.all import hexdump, wrpcap
else:
    from scapy.all import hexdump, PcapNgWriter
    if scapy_mayor > 2 or (scapy_mayor == 2 and scapy_minor >= 6):
        # Support for setting the iface name in the pcapng file was added in:
        # 56b4fa4a pcapng enhancements (idb,epb) and some fixes (#4342)
        scapy_supports_pcap_iface = True

#
# Actual eBPF source code
#
ebpf_source = """
#include <linux/netdevice.h>
#include <linux/skbuff.h>

#define MAX_PACKET <MAX_PACKET_VAL>
#define MAX_KEY    <MAX_KEY_VAL>

struct event_t {
    int result;
    u32 cpu;
    u32 pid;
    u32 upcall_type;
    u64 ts;
    u32 pkt_size;
    u64 key_size;
    char comm[TASK_COMM_LEN];
    char dpif_name[32];
    char dev_name[16];
    unsigned char pkt[MAX_PACKET];
    unsigned char key[MAX_KEY];
};
BPF_RINGBUF_OUTPUT(events, <BUFFER_PAGE_CNT>);
BPF_TABLE("percpu_array", uint32_t, uint64_t, dropcnt, 1);

static
void report_missed_event() {
    uint32_t type = 0;
    uint64_t *value = dropcnt.lookup(&type);
    if (value)
        __sync_fetch_and_add(value, 1);
}

#if <INSTALL_OVS_UPCALL_RECV_PROBE>
int do_trace(struct pt_regs *ctx) {
    uint64_t addr;
    uint64_t size;

    struct event_t *event = events.ringbuf_reserve(sizeof(struct event_t));
    if (!event) {
        report_missed_event();
        return 1;
    }

    event->ts = bpf_ktime_get_ns();
    event->cpu =  bpf_get_smp_processor_id();
    event->pid = bpf_get_current_pid_tgid();
    event->result = 0;
    event->dev_name[0] = 0;
    bpf_get_current_comm(&event->comm, sizeof(event->comm));

    bpf_usdt_readarg(1, ctx, &addr);
    bpf_probe_read_str(&event->dpif_name, sizeof(event->dpif_name),
                       (void *)addr);

    bpf_usdt_readarg(2, ctx, &event->upcall_type);
    bpf_usdt_readarg(4, ctx, &event->pkt_size);
    bpf_usdt_readarg(6, ctx, &event->key_size);

    if (event->pkt_size > MAX_PACKET)
        size = MAX_PACKET;
    else
        size = event->pkt_size;
    bpf_usdt_readarg(3, ctx, &addr);
    bpf_probe_read(&event->pkt, size, (void *)addr);

    if (event->key_size > MAX_KEY)
        size = MAX_KEY;
    else
        size = event->key_size;
    bpf_usdt_readarg(5, ctx, &addr);
    bpf_probe_read(&event->key, size, (void *)addr);

    events.ringbuf_submit(event, 0);
    return 0;
};
#endif

#if <INSTALL_OVS_UPCALL_DROP_PROBE>
struct inflight_upcall {
    u32 cpu;
    u32 upcall_type;
    u64 ts;
    struct sk_buff *skb;
    char dpif_name[32];
};
BPF_HASH(inflight_upcalls, u64, struct inflight_upcall);

TRACEPOINT_PROBE(openvswitch, ovs_dp_upcall)
{
    u64 pid = bpf_get_current_pid_tgid();
    struct inflight_upcall upcall = {};

    upcall.cpu = bpf_get_smp_processor_id();
    upcall.ts = bpf_ktime_get_ns();
    upcall.upcall_type = args->upcall_cmd;
    upcall.skb = args->skbaddr;
    TP_DATA_LOC_READ_CONST(&upcall.dpif_name, dp_name,
                           sizeof(upcall.dpif_name));

    inflight_upcalls.insert(&pid, &upcall);
    return 0;
}

int kretprobe__ovs_dp_upcall(struct pt_regs *ctx)
{
    u64 pid = bpf_get_current_pid_tgid();
    struct inflight_upcall *upcall;
    int ret = PT_REGS_RC(ctx);
    struct net_device *dev;
    u64 size;

    upcall = inflight_upcalls.lookup(&pid);
    inflight_upcalls.delete(&pid);
    if (!upcall)
        return 0;

    /* Successfull upcalls are reported in the USDT probe. */
    if (!ret)
        return 0;

    struct event_t *event = events.ringbuf_reserve(sizeof(struct event_t));
    if (!event) {
        report_missed_event();
        return 1;
    }

    event->ts = upcall->ts;
    event->cpu = upcall->cpu;
    event->pid = pid;
    event->result = ret;
    __builtin_memcpy(&event->dpif_name, &upcall->dpif_name, 32);
    bpf_get_current_comm(&event->comm, sizeof(event->comm));
    event->pkt_size = upcall->skb->len;
    event->upcall_type = upcall->upcall_type;
    event->key_size = 0;
    bpf_probe_read(&dev, sizeof(upcall->skb->dev),
                   ((char *)upcall->skb + offsetof(struct sk_buff, dev)));
    bpf_probe_read_kernel(&event->dev_name, 16, dev->name);

    size = upcall->skb->len - upcall->skb->data_len;
    if (size > MAX_PACKET)
        size = MAX_PACKET;

    bpf_probe_read_kernel(event->pkt, size, upcall->skb->data);
    events.ringbuf_submit(event, 0);
    return 0;
}
#endif
"""

pcap_writer = None


#
# format_key()
#
def format_key(event, decode_dump):
    lines = []
    if event.key_size < options.flow_key_size:
        key_len = event.key_size
    else:
        key_len = options.flow_key_size

    if not key_len:
        return []

    if options.flow_key_decode != 'none':
        lines.append("  Flow key size {} bytes, size captured {} bytes.".
                     format(event.key_size, key_len))

    if options.flow_key_decode == 'hex':
        #
        # Abuse scapy's hex dump to dump flow key
        #
        lines.extend(re.sub('^', ' ' * 4,
            hexdump(
                Ether(bytes(event.key)[:key_len]),
                dump=True),
            flags=re.MULTILINE).split("\n"))

    if options.flow_key_decode == "nlraw":
        lines.extend(decode_dump)

    return lines


#
# print_event()
#
def print_event(ctx, data, size):
    global pcap_writer

    event = b["events"].event(data)
    dp = event.dpif_name.decode("utf-8")

    nla, key_dump = decode_nlm(
        bytes(event.key)[: min(event.key_size, options.flow_key_size)]
    )
    if event.dev_name:
        port = event.dev_name.decode("utf-8")
    elif "OVS_KEY_ATTR_IN_PORT" in nla:
        port_no = struct.unpack("=I", nla["OVS_KEY_ATTR_IN_PORT"])[0]
        port = dp_map.get_port_name(dp.partition("@")[-1], port_no)
        if not port:
            port = str(port_no)
    else:
        port = "Unknown"

    print(
        "{:<18.9f} {:<4} {:<16} {:<10} {:<40} {:<4} {:<10} {:<12} {:<8}".
        format(
            event.ts / 1000000000,
            event.cpu,
            event.comm.decode("utf-8"),
            event.pid,
            "{} ({})".format(port, dp),
            event.upcall_type,
            event.pkt_size,
            event.key_size,
            event.result,
        )
    )

    #
    # Dump flow key information
    #
    key_lines = format_key(event, key_dump)
    for line in key_lines:
        print(line)

    #
    # Decode packet only if there is data
    #
    if event.pkt_size <= 0:
        return

    pkt_id = get_pkt_id()

    if event.pkt_size < options.packet_size:
        pkt_len = event.pkt_size
        pkt_data = bytes(event.pkt)[:event.pkt_size]
    else:
        pkt_len = options.packet_size
        pkt_data = bytes(event.pkt)

    if options.packet_decode != 'none' or options.pcap is not None:
        print("  {}: Receive dp_port {}, packet size {} bytes, size "
              "captured {} bytes.".format(pkt_id, port, event.pkt_size,
                                          pkt_len))

    if options.packet_decode == 'hex':
        print(re.sub('^', ' ' * 4, hexdump(pkt_data, dump=True),
                     flags=re.MULTILINE))

    packet = Ether(pkt_data)
    packet.wirelen = event.pkt_size

    if options.packet_decode == 'decode':
        print(re.sub('^', ' ' * 4, packet.show(dump=True), flags=re.MULTILINE))

    if options.pcap is not None:
        try:
            if pcap_writer is None:
                pcap_writer = PcapNgWriter(options.pcap)

            comment = "cpu={} comm={} pid={} upcall_type={} result={}". format(
                event.cpu, event.comm.decode("utf-8"), event.pid,
                event.upcall_type, event.result)

            if options.flow_key_decode != 'none':
                comment = comment + "\n" + "\n".join(key_lines)

            if scapy_supports_pcap_iface:
                packet.sniffed_on = "{} ({})".format(port, dp)
            else:
                comment = "iface={}({}) ".format(port, dp) + comment

            packet.comment = comment
            pcap_writer.write(packet)
        except NameError:  # PcapNgWriter not found
            wrpcap(options.pcap, packet, append=True,
                   snaplen=options.packet_size)


#
# decode_nlm()
#
def decode_nlm(msg, indent=4):
    bytes_left = len(msg)
    result = {}
    dump = []

    while bytes_left:
        if bytes_left < 4:
            dump.append(
                "{}WARN: decode truncated; can't read header".format(
                    " " * indent
                )
            )
            break

        nla_len, nla_type = struct.unpack("=HH", msg[:4])

        if nla_len < 4:
            dump.append(
                "{}WARN: decode truncated; nla_len < 4".format(" " * indent)
            )
            break

        nla_data = msg[4:nla_len]
        trunc = ""

        if nla_len > bytes_left:
            trunc = "..."
            nla_data = nla_data[:(bytes_left - 4)]
        else:
            result[get_ovs_key_attr_str(nla_type)] = nla_data

        dump.append(
            "{}nla_len {}, nla_type {}[{}], data: {}{}".format(
                ' ' * indent, nla_len, get_ovs_key_attr_str(nla_type),
                nla_type,
                "".join("{:02x} ".format(b) for b in nla_data), trunc)
        )

        if trunc != "":
            dump.append(
                "{}WARN: decode truncated; nla_len > msg_len[{}] ".format(
                    " " * indent, bytes_left
                )
            )
            break

        # update next offset, but make sure it's aligned correctly
        next_offset = (nla_len + 3) & ~(3)
        msg = msg[next_offset:]
        bytes_left -= next_offset

    return result, dump


#
# get_ovs_key_attr_str()
#
def get_ovs_key_attr_str(attr):
    ovs_key_attr = ["OVS_KEY_ATTR_UNSPEC",
                    "OVS_KEY_ATTR_ENCAP",
                    "OVS_KEY_ATTR_PRIORITY",
                    "OVS_KEY_ATTR_IN_PORT",
                    "OVS_KEY_ATTR_ETHERNET",
                    "OVS_KEY_ATTR_VLAN",
                    "OVS_KEY_ATTR_ETHERTYPE",
                    "OVS_KEY_ATTR_IPV4",
                    "OVS_KEY_ATTR_IPV6",
                    "OVS_KEY_ATTR_TCP",
                    "OVS_KEY_ATTR_UDP",
                    "OVS_KEY_ATTR_ICMP",
                    "OVS_KEY_ATTR_ICMPV6",
                    "OVS_KEY_ATTR_ARP",
                    "OVS_KEY_ATTR_ND",
                    "OVS_KEY_ATTR_SKB_MARK",
                    "OVS_KEY_ATTR_TUNNEL",
                    "OVS_KEY_ATTR_SCTP",
                    "OVS_KEY_ATTR_TCP_FLAGS",
                    "OVS_KEY_ATTR_DP_HASH",
                    "OVS_KEY_ATTR_RECIRC_ID",
                    "OVS_KEY_ATTR_MPLS",
                    "OVS_KEY_ATTR_CT_STATE",
                    "OVS_KEY_ATTR_CT_ZONE",
                    "OVS_KEY_ATTR_CT_MARK",
                    "OVS_KEY_ATTR_CT_LABELS",
                    "OVS_KEY_ATTR_CT_ORIG_TUPLE_IPV4",
                    "OVS_KEY_ATTR_CT_ORIG_TUPLE_IPV6",
                    "OVS_KEY_ATTR_NSH"]

    if attr < 0 or attr > len(ovs_key_attr):
        return "<UNKNOWN>"

    return ovs_key_attr[attr]


#
# get_pkt_id()
#
def get_pkt_id():
    if not hasattr(get_pkt_id, "counter"):
        get_pkt_id.counter = 0
    get_pkt_id.counter += 1
    return get_pkt_id.counter


#
# buffer_size_type()
#
def buffer_size_type(astr, min=64, max=2048):
    value = int(astr)
    if min <= value <= max:
        return value
    else:
        raise argparse.ArgumentTypeError(
            'value not in range {}-{}'.format(min, max))


#
# next_power_of_two()
#
def next_power_of_two(val):
    np = 1
    while np < val:
        np *= 2
    return np


#
# main()
#
def main():
    #
    # Don't like these globals, but ctx passing does not seem to work with the
    # existing open_ring_buffer() API :(
    #
    global b
    global options
    global dp_map

    dp_map = DpPortMapping()

    #
    # Argument parsing
    #
    parser = argparse.ArgumentParser()

    parser.add_argument("--buffer-page-count",
                        help="Number of BPF ring buffer pages, default 1024",
                        type=int, default=1024, metavar="NUMBER")
    parser.add_argument("-D", "--debug",
                        help="Enable eBPF debugging",
                        type=int, const=0x3f, default=0, nargs='?')
    parser.add_argument('-d', '--packet-decode',
                        help='Display packet content in selected mode, '
                        'default none',
                        choices=['none', 'hex', 'decode'], default='none')
    parser.add_argument("-f", "--flow-key-size",
                        help="Set maximum flow key size to capture, "
                        "default 64", type=buffer_size_type, default=64,
                        metavar="[64-2048]")
    parser.add_argument('-k', '--flow-key-decode',
                        help='Display flow-key content in selected mode, '
                        'default none',
                        choices=['none', 'hex', 'nlraw'], default='none')
    parser.add_argument("-p", "--pid", metavar="VSWITCHD_PID",
                        help="ovs-vswitch's PID",
                        type=int, default=None)
    parser.add_argument("-s", "--packet-size",
                        help="Set maximum packet size to capture, "
                        "default 64", type=buffer_size_type, default=64,
                        metavar="[64-2048]")
    parser.add_argument("-w", "--pcap", metavar="PCAP_FILE",
                        help="Write upcall packets to specified pcap file.",
                        type=str, default=None)
    parser.add_argument("-r", "--result",
                        help="Display only events with the given result, "
                        "default: any",
                        choices=["error", "ok", "any"], default="any")

    options = parser.parse_args()

    #
    # Check if current kernel supports error reporting.
    #
    tracefs_paths = ("/sys/kernel/debug/tracing/", "/sys/kernel/tracing/")
    upcall_tp_found = False
    for tracefs in tracefs_paths:
        if exists(join(tracefs, "events/openvswitch/ovs_dp_upcall")):
            upcall_tp_found = True
            break

    if not upcall_tp_found:
        if options.result == "error":
            print("ERROR: Monitoring error upcalls is not supported by the "
                  "running kernel (or the tracefs is not mounted).")
            sys.exit(-1)
        if options.result == "any":
            print("WARN: Monitoring error upcalls is not supported by the "
                  "running kernel (or the tracefs is not mounted). "
                  "Only successful ones will be monitored.")
            options.result = "ok"

    #
    # Find the PID of the ovs-vswitchd daemon if not specified.
    #
    if options.pid is None:
        for proc in psutil.process_iter():
            if 'ovs-vswitchd' in proc.name():
                if options.pid is not None:
                    print("ERROR: Multiple ovs-vswitchd daemons running, "
                          "use the -p option!")
                    sys.exit(-1)

                options.pid = proc.pid

    #
    # Error checking on input parameters
    #
    if options.pid is None:
        print("ERROR: Failed to find ovs-vswitchd's PID!")
        sys.exit(-1)

    if options.pcap is not None:
        if exists(options.pcap):
            print("ERROR: Destination capture file \"{}\" already exists!".
                  format(options.pcap))
            sys.exit(-1)

    options.buffer_page_count = next_power_of_two(options.buffer_page_count)

    #
    # Attach the usdt probe
    #
    usdt = []
    if options.result in ["ok", "any"]:
        u = USDT(pid=int(options.pid))
        try:
            u.enable_probe(probe="recv_upcall", fn_name="do_trace")
            usdt.append(u)
        except USDTException as e:
            print("ERROR: {}"
                  "ovs-vswitchd!".format(
                      (re.sub('^', ' ' * 7, str(e), flags=re.MULTILINE)).
                      strip().replace("--with-dtrace or --enable-dtrace",
                                      "--enable-usdt-probes")))
            sys.exit(-1)

    #
    # Uncomment to see how arguments are decoded.
    #       print(u.get_text())
    #

    #
    # Attach probe to running process
    #
    source = ebpf_source.replace("<MAX_PACKET_VAL>", str(options.packet_size))
    source = source.replace("<MAX_KEY_VAL>", str(options.flow_key_size))
    source = source.replace("<BUFFER_PAGE_CNT>",
                            str(options.buffer_page_count))
    source = source.replace("<INSTALL_OVS_UPCALL_RECV_PROBE>", "1"
                            if options.result in ["ok", "any"] else "0")
    source = source.replace("<INSTALL_OVS_UPCALL_DROP_PROBE>", "1"
                            if options.result in ["error", "any"] else "0")

    b = BPF(text=source, usdt_contexts=usdt, debug=options.debug)

    #
    # Print header
    #
    print("{:<18} {:<4} {:<16} {:<10} {:<40} {:<4} {:<10} {:<12} {:<8}".format(
        "TIME", "CPU", "COMM", "PID", "PORT_NAME", "TYPE", "PKT_LEN",
        "FLOW_KEY_LEN", "RESULT"))

    #
    # Dump out all events
    #
    b['events'].open_ring_buffer(print_event)
    while 1:
        try:
            b.ring_buffer_poll()
        except KeyboardInterrupt:
            break

    dropcnt = b.get_table("dropcnt")
    for k in dropcnt.keys():
        count = dropcnt.sum(k).value
        if k.value == 0 and count > 0:
            print("\nWARNING: Not all upcalls were captured, {} were dropped!"
                  "\n         Increase the BPF ring buffer size with the "
                  "--buffer-page-count option.".format(count))


#
# Start main() as the default entry point...
#
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# Copyright (c) 2022-2024 Redhat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Script information:
# -------------------
# flow_reval_monitor.py uses the dpif_netlink_operate:flow_put and
# revalidator:flow_result USDT probes to monitor flow lifetimes and
# expiration events. By default, this will show all flow_put and flow
# expiration events, along with their reasons. This will look like so:
#
# TID        TIME               UFID                               EVENT/REASON
# 71828 1549.119959874 39f0f28f-33... Insert (put) flow to ovs kernel module.
# 71828 1549.420877223 850db41c-47... Insert (put) flow to ovs kernel module.
# 71828 1550.476923456 5bacfca9-fe... Insert (put) flow to ovs kernel module.
# 71832      1559.650192299     850db41c-47... Idle flow timed out
# 71832      1561.153332825     39f0f28f-33... Idle flow timed out
# 71832      1572.684316304     5bacfca9-fe... Idle flow timed out
#
# Flow key data can be printed using the --flow-keys option.  This will
# print the equivalent datapath flow string.
#
# When filtering flows, the syntax is the same as used by
# `ovs-appctl dpctl/add-flow`.
#
# For a complete list of options, please use the '--help' or '-h' argument.
#
# Examples:
#
# To use the script on a running ovs-vswitchd to see flow keys and expiration
# events for flows with an ipv4 source of 192.168.10.10:
# $ ./flow_reval_monitor.py --flow-keys --filter-flows \
#   "ipv4(src=192.168.10.10)"
# TIME               UFID                                          EVENT/REASON
# 105082.457322742   ufid:f76fc899-376d-466b-bc74-0000b933eb97     flow_put
# ufid:f76fc899-376d-466b-bc74-0000b933eb97 has the following flow information:
#     in_port(2),
#     eth(src=0e:04:47:fc:74:51, dst=da:dc:c5:69:05:d7), \
#     eth_type(0x800), \
#     ipv4(src=192.168.10.10, dst=192.168.10.30, proto=1, tos=0, ttl=64,[...]),
#     icmp(type=8, code=0)
# 105092.635450202   ufid:f76fc899-376d-466b-bc74-0000b933eb97   Flow timed out
#
# Notes:
#   1) No options are needed to attach when there is a single running instance
#      of ovs-vswitchd.
#   2) If you're using the flow filtering option, it will only track flows that
#      have been upcalled since the script began running.
#   3) When using the flow filtering option, the key size will likely need to
#      be expanded to match on all the fields in the message.  The default is
#      kept small to keep the buffer copy sizes down when displaying
#      flows (-k), but is hardcoded to 2048 when an actual filter (-l) is
#      applied
#   4) The flow filtering format is a simplified form of the ODP syntax, and
#      does not support masked matches, which means you will need to filter
#      on exact details.  The fields present are dependent on how the
#      classifier and OFP rules form the ODP rules - not all fields may be
#      present in a particular flow.
#   5) The flow_put filtering only happens for flows installed into the ovs
#      kernel module.  This means flows taking the HW offload path (ie: tc),
#      or on DPDK side won't get matched.

try:
    from bcc import BPF
    from bcc import USDT
    from bcc import USDTException
except ModuleNotFoundError:
    print("ERROR: Can't find the BPF Compiler Collection Tools.")
    print("Please install them before running this script.")
    exit(1)

from enum import IntEnum
from ipaddress import IPv4Address, IPv6Address
from pathlib import Path

import argparse
import psutil
import re
import struct
import subprocess
import sys

#
# eBPF source code
#
bpf_src = """
#include <linux/sched.h>

#define MAX_KEY      <MAX_KEY_VAL>
#define FLOW_FILTER  <FILTER_BOOL>

enum probe { <EVENT_ENUM> };

<OVS_INCLUDE_DEFINITIONS>

struct event_t {
    u64 ts;
    u32 pid;
    u32 result;
    u32 reason;
    u32 ufid[4];
    u64 key_size;
    unsigned char key[MAX_KEY];
    enum probe probe;
};

BPF_HASH(watchlist, ovs_u128);
BPF_RINGBUF_OUTPUT(events, <BUFFER_PAGE_COUNT>);
BPF_TABLE("percpu_array", uint32_t, uint64_t, dropcnt, 1);

/* Hack to make a 'static' like storage object. */
BPF_TABLE("percpu_array", uint32_t, struct udpif_key, udpk, 1);

static struct event_t *get_event(enum probe p) {
    struct event_t *event = events.ringbuf_reserve(sizeof(struct event_t));

    if (!event) {
        dropcnt.increment(0);
        return NULL;
    }

    event->probe = p;
    event->ts = bpf_ktime_get_ns();
    event->pid = bpf_get_current_pid_tgid();

    return event;
}

static int emit_flow_result(struct udpif_key *ukey, ovs_u128 ufid,
                            u32 result, u32 reason) {
    struct event_t *event = NULL;
    u64 *ufid_present = NULL;

    ufid_present = watchlist.lookup(&ufid);
    if (FLOW_FILTER && !ufid_present) {
        return 0;
    }

    event = get_event(FLOW_RESULT);
    if (!event) {
        /* If we can't reserve the space in the ring buffer, return 1. */
        return 1;
    }

    event->result = result;
    event->reason = reason;
    bpf_probe_read(&event->ufid, sizeof ufid, &ufid);
    events.ringbuf_submit(event, 0);

    return 0;
}

int usdt__flow_result(struct pt_regs *ctx) {
    struct udpif_key *ukey = NULL;
    u32 reason = 0;
    u32 result = 0;
    ovs_u128 ufid;
    u32 zero = 0;

    ukey = udpk.lookup(&zero);
    if (!ukey) {
        return 1;
    }
    bpf_usdt_readarg_p(2, ctx, ukey, sizeof(struct udpif_key));
    bpf_usdt_readarg(3, ctx, &result);
    bpf_usdt_readarg(4, ctx, &reason);
    ufid = ukey->ufid;

    return emit_flow_result(ukey, ufid, result, reason);
}

int usdt__flow_sweep_result(struct pt_regs *ctx) {
    struct udpif_key *ukey = NULL;
    u32 reason = 0;
    u32 result = 0;
    ovs_u128 ufid;
    u32 zero = 0;

    ukey = udpk.lookup(&zero);
    if (!ukey) {
        return 1;
    }
    bpf_usdt_readarg_p(2, ctx, ukey, sizeof(struct udpif_key));
    bpf_usdt_readarg(3, ctx, &result);
    bpf_usdt_readarg(4, ctx, &reason);
    ufid = ukey->ufid;

    return emit_flow_result(ukey, ufid, result, reason);
}

int usdt__op_flow_put(struct pt_regs *ctx) {
    struct dpif_flow_put put;
    ovs_u128 ufid;

    struct event_t *event = get_event(OP_FLOW_PUT);
    if (!event) {
        /* If we can't reserve the space in the ring buffer, return 1. */
        return 1;
    }

    bpf_usdt_readarg_p(2, ctx, &put, sizeof put);
    bpf_probe_read(&event->ufid, sizeof event->ufid, put.ufid);
    bpf_probe_read(&ufid, sizeof ufid, &event->ufid);
    if (put.key_len > MAX_KEY) {
        put.key_len = MAX_KEY;
    }
    event->key_size = put.key_len;
    bpf_probe_read(&event->key, put.key_len, put.key);
    event->reason = 0;
    events.ringbuf_submit(event, 0);

    watchlist.increment(ufid);
    return 0;
}
"""

Event = IntEnum("Event", ["OP_FLOW_PUT", "FLOW_RESULT"], start=0)
RevalResult = IntEnum(
    "reval_result",
    [
        "UKEY_KEEP",
        "UKEY_DELETE",
        "UKEY_MODIFY",
    ],
    start=0,
)

#
# The below FdrReasons and FdrReasonStrings definitions can be found in the
# ofproto/ofproto-dpif-upcall.c file.  Please keep them in sync.
#
FdrReasons = IntEnum(
    "flow_del_reason",
    [
        "FDR_NONE",
        "FDR_AVOID_CACHING",
        "FDR_BAD_ODP_FIT",
        "FDR_FLOW_IDLE",
        "FDR_FLOW_LIMIT",
        "FDR_FLOW_WILDCARDED",
        "FDR_NO_OFPROTO",
        "FDR_PURGE",
        "FDR_TOO_EXPENSIVE",
        "FDR_UPDATE_FAIL",
        "FDR_XLATION_ERROR",
        "FDR_FLOW_MISSING_DP"
    ],
    start=0,
)

FdrReasonStrings = {
    FdrReasons.FDR_NONE: "No delete reason specified",
    FdrReasons.FDR_AVOID_CACHING: "Cache avoidance flag set",
    FdrReasons.FDR_BAD_ODP_FIT: "Bad ODP flow fit",
    FdrReasons.FDR_FLOW_IDLE: "Flow idle timeout",
    FdrReasons.FDR_FLOW_LIMIT: "Kill all flows condition reached",
    FdrReasons.FDR_FLOW_WILDCARDED: "Flow needs a narrower wildcard mask",
    FdrReasons.FDR_NO_OFPROTO: "Bridge not found",
    FdrReasons.FDR_PURGE: "User requested flow deletion",
    FdrReasons.FDR_TOO_EXPENSIVE: "Too expensive to revalidate",
    FdrReasons.FDR_UPDATE_FAIL: "Datapath update failed",
    FdrReasons.FDR_XLATION_ERROR: "Flow translation error",
    FdrReasons.FDR_FLOW_MISSING_DP: "Flow is missing from the datapath"
}


def err(msg, code=-1):
    """Prints an error to stderr and exits"""

    print(msg, file=sys.stderr)
    sys.exit(code)


def run_program(command):
    """Invokes a new process and returns stdout.  Note that this will honor
    the PATH environment variable, so best to use it sparingly, or with a
    full path to binary."""

    try:
        process = subprocess.run(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="utf8",
            check=True,
        )

    except subprocess.CalledProcessError as perror:
        return perror.returncode, perror.stdout

    return 0, process.stdout


def get_ovs_definitions(objects, pahole="pahole", pid=None):
    """Uses `pahole` or similar utility to pull object definitions from a
    running OVS process.  The objects argument can either be a string
    or can be a list of strings.  Optionally, pass a specific `pahole`
    binary to use rather than the default.  PID needs to be set."""

    if pid is None:
        raise ValueError("A valid pid value should be supplied!")

    if not isinstance(objects, list):
        objects = [objects]

    if len(objects) == 0:
        raise ValueError("Must supply at least one object!")

    vswitchd = Path(f"/proc/{pid}/exe").resolve()

    object_str = ",".join(objects)

    def run_pahole(debug_file):
        """Helper designed for running pahole, or something with compatible
        output"""

        error, result = run_program(
            [pahole, "-C", object_str, "--compile", debug_file]
        )

        if error:
            if f"pahole: {debug_file}: Invalid argument" not in result:
                err(
                    "ERROR: Pahole failed to get ovs-vswitchd data "
                    "structures!\n{}".format(
                        re.sub(
                            "^", " " * 7, result.rstrip(), flags=re.MULTILINE
                        )
                    )
                )

            return None

        if bool(re.search("pahole: type .* not found", result)):
            return None

        return result

    def run_readelf(bin_file):
        """Helper designed for running readelf or something with compatible
        output"""

        error, result = run_program(
            ["readelf", "-n", "--debug-dump=links", bin_file]
        )

        if error:
            err(
                "ERROR: Failed 'readelf' on \"{}\"!\n{}".format(
                    bin_file, re.sub("^", " " * 7, result, flags=re.MULTILINE)
                )
            )

        return result

    def get_debug_file(bin_file):
        """Runs readelf against the binary, and attempts to find the associated
        debuginfo file."""
        elf_result = run_readelf(bin_file)
        match = re.search("Build ID: ([0-9a-fA-F]+)", elf_result)
        if not match:
            err("ERROR: Can't find build ID to read debug symbols!")

        dbg_file = "/usr/lib/debug/.build-id/{}/{}.debug".format(
            match.group(1)[:2], match.group(1)[2:]
        )

        return dbg_file

    def get_from_shared_library(debug_file):
        ovs_libs = [
            "libofproto",
            "libopenvswitch",
            "libovsdb",
            "libsflow",
            "libvtep",
        ]
        error, ldd_result = run_program(["ldd", debug_file])

        if error:
            err(
                "ERROR: Failed 'ldd' on \"{}\"!\n{}".format(
                    debug_file,
                    re.sub("^", " " * 7, ldd_result, flags=re.MULTILINE),
                )
            )

        for lib in ovs_libs:
            match = re.search(
                r"^\s*{}.* => (.*) \(.*\)$".format(lib),
                ldd_result,
                flags=re.MULTILINE,
            )
            if match is None:
                continue

            result = run_pahole(match.group(1))
            if result is None:
                result = run_pahole(get_debug_file(match.group(1)))

            if result:
                return result

        return None

    #
    # First try to find the debug data as part of the executable.
    #
    result = run_pahole(vswitchd)

    if result is None:
        print(f'INFO: Failed to find debug info in "{vswitchd}"!')

        #
        # Get additional .debug information if available.
        #
        dbg_file = get_debug_file(vswitchd)
        result = run_pahole(dbg_file)
        if result is None:
            print(f'INFO: Failed to find debug info in "{dbg_file}"!')

        #
        # Try to get information from shared libraries if used.
        #
        result = get_from_shared_library(vswitchd)

    if result is None:
        err(f"ERROR: Failed to find needed data structures through {pahole}")

    #
    # We need an empty _Atomic definition to avoid compiler complaints.
    #
    result = "#define _Atomic\n" + result

    #
    # Remove the uint64_t definition as it conflicts with the kernel one.
    #
    result = re.sub("^typedef.*uint64_t;$", "", result, flags=re.MULTILINE)

    return result


def buffer_size_type(astr, min=64, max=2048):
    """Checks whether a string passed in is a number between min and max."""

    value = int(astr)
    if min <= value <= max:
        return value
    else:
        raise argparse.ArgumentTypeError(
            "value not in range {}-{}".format(min, max)
        )


def format_ufid(ufid):
    """Formats a UFID object into a human readable form.  If ufid is None,
    prints "ufid:none" instead."""
    if ufid is None:
        return "ufid:none"

    return "{:08x}-{:04x}-{:04x}-{:04x}-{:04x}{:08x}".format(
        ufid[0],
        ufid[1] >> 16,
        ufid[1] & 0xFFFF,
        ufid[2] >> 16,
        ufid[2] & 0,
        ufid[3],
    )


def find_and_delete_from_watchlist(event):
    """If the event ufid is in the watchlist, delete it"""

    for k, _ in b["watchlist"].items():
        key_ufid = struct.unpack("=IIII", k)
        if key_ufid == tuple(event.ufid):
            key = (b["watchlist"].Key * 1)(k)
            b["watchlist"].items_delete_batch(key)
            break


def handle_flow_put(event):
    """Event handler for the `flow_put` action.  This function will try
    to populate the watchlist based on the vswitchd emitting a put event
    to push an ODP flow key with associated actions into the kernel module"""

    if args.flow_keys or args.filter_flows is not None:
        key = decode_key(bytes(event.key)[: event.key_size])
        flow_dict, flow_str = parse_flow_dict(key)
        # For each attribute that we're watching.
        if args.filter_flows is not None:
            if not compare_flow_to_target(args.filter_flows, flow_dict):
                find_and_delete_from_watchlist(event)
                return

    print(
        "{:<10} {:<18.9f} {:<36} {}".format(
            event.pid,
            event.ts / 1000000000,
            format_ufid(event.ufid),
            "Insert (put) flow to ovs kernel module.",
        )
    )

    if args.flow_keys and len(flow_str):
        flow_str_fields = flow_str.split("), ")
        flow_str = "    "
        curlen = 4
        for field in flow_str_fields:
            if curlen + len(field) > 79:
                flow_str += "\n    "
                curlen = 4
            if field[-1] != ")":
                field += ")"
            flow_str += field + ", "
            curlen += len(field) + 2

        print(" - It holds the following key information:")
        print(flow_str)


def compare_flow_to_target(target, flow):
    """Routine to compare two flow keys"""

    for key in target:
        if key not in flow:
            return False
        elif target[key] is True:
            continue
        elif target[key] == flow[key]:
            continue
        elif isinstance(target[key], dict) and isinstance(flow[key], dict):
            return compare_flow_to_target(target[key], flow[key])
        else:
            return False
    return True


#
# parse_flow_str()
#
def parse_flow_str(flow_str):
    """Loosely parses an ODP flow key into a dict for further processing"""

    f_list = [i.strip(", ") for i in flow_str.split(")")]
    if f_list[-1] == "":
        f_list = f_list[:-1]
    flow_dict = {}
    for e in f_list:
        split_list = e.split("(")
        k = split_list[0]
        if len(split_list) == 1:
            flow_dict[k] = True
        elif split_list[1].count("=") == 0:
            flow_dict[k] = split_list[1]
        else:
            sub_dict = {}
            sublist = [i.strip() for i in split_list[1].split(",")]
            for subkey in sublist:
                brk = subkey.find("=")
                sub_dict[subkey[:brk]] = subkey[brk + 1 :]
            flow_dict[k] = sub_dict
    return flow_dict


def print_expiration(event):
    """Prints a UFID eviction with a reason."""
    ufid_str = format_ufid(event.ufid)

    try:
        reason = FdrReasonStrings[event.reason]
    except KeyError:
        reason = f"Unknown reason '{event.reason}'"

    print(
        "{:<10} {:<18.9f} {:<36} {:<17}".format(
            event.pid,
            event.ts / 1000000000,
            ufid_str,
            reason,
        )
    )


def decode_key(msg):
    """Decodes netlink OVS key attribute."""
    bytes_left = len(msg)
    result = {}
    while bytes_left:
        if bytes_left < 4:
            break
        nla_len, nla_type = struct.unpack("=HH", msg[:4])
        if nla_len < 4:
            break
        nla_data = msg[4:nla_len]
        if nla_len > bytes_left:
            nla_data = nla_data[: (bytes_left - 4)]
            break
        else:
            result[get_ovs_key_attr_str(nla_type)] = nla_data
        next_offset = (nla_len + 3) & (~3)
        msg = msg[next_offset:]
        bytes_left -= next_offset
    if bytes_left:
        print(f"INFO: Buffer truncated with {bytes_left} bytes left.")
    return result


#
# get_ovs_key_attr_str()
#
def get_ovs_key_attr_str(attr):
    ovs_key_attr = [
        "OVS_KEY_ATTR_UNSPEC",
        "encap",
        "skb_priority",
        "in_port",
        "eth",
        "vlan",
        "eth_type",
        "ipv4",
        "ipv6",
        "tcp",
        "udp",
        "icmp",
        "icmpv6",
        "arp",
        "nd",
        "skb_mark",
        "tunnel",
        "sctp",
        "tcp_flags",
        "dp_hash",
        "recirc_id",
        "mpls",
        "ct_state",
        "ct_zone",
        "ct_mark",
        "ct_label",
        "ct_tuple4",
        "ct_tuple6",
        "nsh",
    ]

    if attr < 0 or attr > len(ovs_key_attr):
        return "<UNKNOWN>: {}".format(attr)
    return ovs_key_attr[attr]


def parse_flow_dict(key_dict, decode=True):
    """Processes a flow key dict (see `parse_flow_str` or `decode_key`) and
    returns a tuple of both the final flow key dict, and a string that
    represents and ODP-like representation.  Attempts to decode the actual
    data values if `decode` is true.  Otherwise, this can be for a loose form
    of validation.  Throws a KeyError when it encounters an unknown flow
    key."""

    ret_str = ""
    parseable = {}
    skip = ["nsh", "tunnel", "mpls", "vlan"]
    need_byte_swap = ["ct_label"]
    ipv4addrs = ["ct_tuple4", "tunnel", "ipv4", "arp"]
    ipv6addrs = ["ipv6", "nd", "ct_tuple6"]
    macs = {"eth": [0, 1], "arp": [3, 4], "nd": [1, 2]}
    fields = [
        ("OVS_KEY_ATTR_UNSPEC"),
        ("encap",),
        ("skb_priority", "<I"),
        ("in_port", "<I"),
        ("eth", "!6s6s", "src", "dst"),
        ("vlan",),
        ("eth_type", "!H"),
        ("ipv4", "!4s4s4B", "src", "dst", "proto", "tos", "ttl", "frag"),
        (
            "ipv6",
            "!16s16s4s4B",
            "src",
            "dst",
            "label",
            "proto",
            "tclass",
            "hlimit",
            "frag",
        ),
        ("tcp", "!2H", "src", "dst"),
        ("udp", "!2H", "src", "dst"),
        ("icmp", "!2B", "type", "code"),
        ("icmpv6", "!2B", "type", "code"),
        ("arp", "!4s4sH6s6s", "sip", "tip", "op", "sha", "tha"),
        ("nd", "!16s6s6s", "target", "sll", "tll"),
        ("skb_mark", "<I"),
        ("tunnel",),
        ("sctp", "!2H", "src", "dst"),
        ("tcp_flags", "!H"),
        ("dp_hash", "<I"),
        ("recirc_id", "<I"),
        ("mpls",),
        ("ct_state", "<I"),
        ("ct_zone", "<H"),
        ("ct_mark", "<I"),
        ("ct_label", "!16s"),
        ("ct_tuple4", "!4s4s2HB", "src", "dst", "tp_src", "tp_dst", "proto"),
        ("ct_tuple6", "!16s16sB2H", "src", "dst", "proto", "tp_src", "tp_dst"),
        ("nsh",),
    ]
    for k, v in key_dict.items():
        s = ""
        if k in skip:
            continue
        if decode and int.from_bytes(v, "big") == 0:
            parseable[k] = "0"
            continue
        if decode and k in need_byte_swap:
            v = int.from_bytes(v, "little").to_bytes(len(v), "big")
        attr = -1
        found = False
        for f in fields:
            if k == f[0]:
                attr = fields.index(f)
                found = True
                break
        if not found:
            raise KeyError("Invalid flow field '%s'" % k)
        if decode and len(fields[attr]) > 1:
            data = list(
                struct.unpack(
                    fields[attr][1], v[: struct.calcsize(fields[attr][1])]
                )
            )
            if k in ipv4addrs:
                if data[0].count(0) < 4:
                    data[0] = str(IPv4Address(data[0]))
                else:
                    data[0] = b"\x00"
                if data[1].count(0) < 4:
                    data[1] = str(IPv4Address(data[1]))
                else:
                    data[1] = b"\x00"
            if k in ipv6addrs:
                if data[0].count(0) < 16:
                    data[0] = str(IPv6Address(data[0]))
                else:
                    data[0] = b"\x00"
                if data[1].count(0) < len(data[1]):
                    data[1] = str(IPv6Address(data[1]))
                else:
                    data[1] = b"\x00"
            if k in macs.keys():
                for e in macs[k]:
                    if data[e].count(0) == 6:
                        mac_str = b"\x00"
                    else:
                        mac_str = ":".join(["%02x" % i for i in data[e]])
                    data[e] = mac_str
        if decode and len(fields[attr]) > 2:
            field_dict = dict(zip(fields[attr][2:], data))
            s = ", ".join(k + "=" + str(v) for k, v in field_dict.items())
        elif decode and k != "eth_type":
            s = str(data[0])
            field_dict = s
        else:
            if decode:
                s = hex(data[0])
            field_dict = s
        ret_str += k + "(" + s + "), "
        parseable[k] = field_dict
    ret_str = ret_str[:-2]
    return (parseable, ret_str)


def handle_event(ctx, data, size):
    """Dispatches to the correct event handler based on the event probe
    type.

    Once we grab the event, we have three cases.
      1. It's a revalidator probe and the reason is nonzero: A flow is expiring
      2. It's a revalidator probe and the reason is zero: flow revalidated
      3. It's a flow_put probe.

    We will ignore case 2, and report all others.
    """

    event = b["events"].event(data)
    if event.probe == Event.OP_FLOW_PUT:
        handle_flow_put(event)
    elif (
        event.probe == Event.FLOW_RESULT
        and event.result == RevalResult.UKEY_DELETE
    ):
        print_expiration(event)


def main():
    #
    # Don't like these globals, but ctx passing does not work with the existing
    # open_ring_buffer() API :(
    #
    global b
    global args

    #
    # Argument parsing
    #
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--buffer-page-count",
        help="Number of BPF ring buffer pages, default 1024",
        type=int,
        default=1024,
        metavar="NUMBER",
    )
    parser.add_argument(
        "-f",
        "--flow-key-size",
        help="Set maximum flow key size to capture, "
        "default 128 - see notes",
        type=buffer_size_type,
        default=128,
        metavar="[128-2048]",
    )
    parser.add_argument(
        "-k",
        "--flow-keys",
        help="Print flow keys as flow strings",
        action="store_true",
    )
    parser.add_argument(
        "-l",
        "--filter-flows",
        metavar="FLOW_STRING",
        help="Filter flows that match the specified " "ODP-like flow",
        type=str,
        default=None,
        nargs="*",
    )
    parser.add_argument(
        "-P",
        "--pahole",
        metavar="PAHOLE",
        help="Pahole executable to use, default pahole",
        type=str,
        default="pahole",
    )
    parser.add_argument(
        "-p",
        "--pid",
        metavar="VSWITCHD_PID",
        help="ovs-vswitchd's PID",
        type=int,
        default=None,
    )
    parser.add_argument(
        "-D",
        "--debug",
        help="Enable eBPF debugging",
        type=int,
        const=0x3F,
        default=0,
        nargs="?",
    )
    args = parser.parse_args()

    #
    # Find the PID of the ovs-vswitchd daemon if not specified.
    #
    if args.pid is None:
        for proc in psutil.process_iter():
            if "ovs-vswitchd" in proc.name():
                if args.pid is not None:
                    err(
                        "Error: Multiple ovs-vswitchd daemons running, "
                        "use the -p option!"
                    )

                args.pid = proc.pid
    #
    # Error checking on input parameters
    #
    if args.pid is None:
        err("ERROR: Failed to find ovs-vswitchd's PID!")

    #
    # Attach the USDT probes
    #
    try:
        u = USDT(pid=int(args.pid))
        u.enable_probe(probe="op_flow_put", fn_name="usdt__op_flow_put")
        u.enable_probe(probe="flow_result", fn_name="usdt__flow_result")
        u.enable_probe(
            probe="flow_sweep_result", fn_name="usdt__flow_sweep_result"
        )
    except USDTException as e:
        err("Failed to attach probes due to:\n" + str(e))

    #
    # Attach the probes to the running process
    #
    source = bpf_src.replace(
        "<BUFFER_PAGE_COUNT>", str(args.buffer_page_count)
    )

    source = source.replace(
        "<OVS_INCLUDE_DEFINITIONS>",
        get_ovs_definitions(
            ["udpif_key", "ovs_u128", "dpif_flow_put"],
            pid=args.pid,
            pahole=args.pahole,
        ),
    )

    if args.filter_flows is None:
        filter_bool = 0

        # Set the key size based on what the user wanted
        source = source.replace("<MAX_KEY_VAL>", str(args.flow_key_size))
    else:
        filter_bool = 1
        args.filter_flows = parse_flow_str(args.filter_flows[0])

        # Run through the parser to make sure we only filter on fields we
        # understand
        parse_flow_dict(args.filter_flows, False)

        # This is hardcoded here because it doesn't make sense to shrink the
        # size, since the flow key might be missing fields that are matched in
        # the flow filter.
        source = source.replace("<MAX_KEY_VAL>", "2048")

    source = source.replace("<FILTER_BOOL>", str(filter_bool))

    source = source.replace(
        "<EVENT_ENUM>",
        "\n".join([f"{event.name} = {event.value}," for event in Event]),
    )

    b = BPF(text=source, usdt_contexts=[u], debug=args.debug)

    #
    # Print header
    #
    print(
        "{:<10} {:<18} {:<36} {:<17}".format(
            "TID", "TIME", "UFID", "EVENT/REASON"
        )
    )

    #
    # Dump out all events.
    #
    b["events"].open_ring_buffer(handle_event)
    while 1:
        try:
            b.ring_buffer_poll()
        except KeyboardInterrupt:
            break

    dropcnt = b.get_table("dropcnt")
    for k in dropcnt.keys():
        count = dropcnt.sum(k).value
        if k.value == 0 and count > 0:
            print(
                "\n# WARNING: Not all flow operations were captured, {} were"
                " dropped!\n#          Increase the BPF ring buffer size "
                "with the --buffer-page-count option.".format(count)
            )


#
# Start main() as the default entry point
#
if __name__ == "__main__":
    main()



# Copyright (c) 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import enum
import functools
import ipaddress
import os
import time
import typing

try:
    import unbound  # type: ignore
except ImportError:
    pass

import ovs.vlog

vlog = ovs.vlog.Vlog("dns_resolve")

RESOLVE_UNUSED_TIMEOUT = 300  # Seconds.


class ReqState(enum.Enum):
    INVALID = 0
    PENDING = 1
    GOOD_UNUSED = 2
    GOOD = 3
    ERROR = 4


class DNSRequest:
    def __init__(self, name: str):
        self.name: str = name
        self.state: ReqState = ReqState.INVALID
        self.time: typing.Optional[float] = None
        # set by DNSResolver._callback
        self.result: typing.Optional[str] = None
        self.ttl: typing.Optional[float] = None

    @property
    def expired(self):
        ttl = self.ttl
        # When we just sent a request, but didn't look at the response yet,
        # it's not caching, but a "transaction in progress" situation, so we
        # can use the response even with TTL of 0 and more than 1 second
        # passed.  Allow such values to be accessed for at least
        # RESOLVE_UNUSED_TIMEOUT seconds without considering them stale.
        # This is necessary in case of large backoff intervals on connections
        # or if the process is doing some other work not looking at the
        # response for longer than TTL.
        if self.state == ReqState.GOOD_UNUSED:
            ttl = max(ttl, RESOLVE_UNUSED_TIMEOUT)
            # Not a "transaction in progress" anymore, normal caching rules
            # should apply from this point forward.
            self.state = ReqState.GOOD

        return time.time() > self.time + ttl

    @property
    def is_valid(self):
        return (self.state in [ReqState.GOOD_UNUSED, ReqState.GOOD]
                and not self.expired)

    def __str__(self):
        return (f"DNSRequest(name={self.name}, state={self.state}, "
                f"time={self.time}, result={self.result})")


class DefaultReqDict(collections.defaultdict):
    def __init__(self):
        super().__init__(DNSRequest)

    def __missing__(self, key):
        ret = self.default_factory(key)
        self[key] = ret
        return ret


class UnboundException(Exception):
    def __init__(self, message, errno):
        try:
            msg = f"{message}: {unbound.ub_strerror(errno)}"
        except NameError:
            msg = message
        super().__init__(msg)


def dns_enabled(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.dns_enabled:
            return func(self, *args, **kwargs)
        vlog.err("DNS support requires the python unbound library")
    return wrapper


class DNSResolver:
    def __init__(self, is_daemon: bool = False):
        """Create a resolver instance

        If is_daemon is true, set the resolver to handle requests
        asynchronously. The following environment variables are processed:

        OVS_UNBOUND_CONF: The filename for an unbound.conf file
        OVS_RESOLV_CONF: A filename to override the system default resolv.conf
        OVS_HOSTS_FILE: A filename to override the system default hosts file

        In the event that the unbound library is missing or fails to initialize
        DNS lookup support will be disabled and the resolve() method will
        return None.
        """
        self._is_daemon = is_daemon
        try:
            self._ctx = unbound.ub_ctx()
            self.dns_enabled = True
        except Exception:
            # The unbound docs mention that this could thrown an exception
            # but do not specify what exception that is. This can also
            # happen with a missing unbound library.
            self.dns_enabled = False
            vlog.err("Failed to initialize the unbound library")
            return

        # NOTE(twilson) This cache, like the C version, can grow without bound
        # and has no cleanup or aging mechanism. Given our usage patterns, this
        # should not be a problem. But this should not be used to resolve an
        # unbounded list of addresses in a long-running daemon.
        self._requests = DefaultReqDict()

        self._ub_call(self._set_unbound_conf)

        # NOTE(twilson) The C version disables DNS in this case. I didn't do
        # that here since it could still be useful to resolve addresses from
        # /etc/hosts even w/o resolv.conf
        self._ub_call(self._set_resolv_conf)
        self._ub_call(self._set_hosts_file)

        self._ctx.set_async(True)  # Sets threaded behavior for resolve_async()

    def _ub_call(self, fn, *args, **kwargs):
        """Convert UnboundExceptions into vlog warnings"""
        try:
            return fn(*args, **kwargs)
        except UnboundException as e:
            vlog.warn(e)

    @dns_enabled
    def _set_unbound_conf(self):
        ub_cfg = os.getenv("OVS_UNBOUND_CONF")
        if ub_cfg:
            retval = self._ctx.config(ub_cfg)
            if retval != 0:
                raise UnboundException(
                    "Failed to set libunbound context config", retval)

    @dns_enabled
    def _set_resolv_conf(self):
        filename = os.getenv("OVS_RESOLV_CONF")
        # The C lib checks that the file exists and also sets filename to
        # /etc/resolv.conf on non-Windows, but resolvconf already does this.
        retval = self._ctx.resolvconf(filename)
        if retval != 0:
            location = filename or "system default nameserver"
            raise UnboundException(location, retval)

    @dns_enabled
    def _set_hosts_file(self):
        # The C lib doesn't have the ability to set a hosts file, but it is
        # useful to have, especially for writing tests that don't rely on
        # network connectivity. hosts(None) uses /etc/hosts.
        filename = os.getenv("OVS_HOSTS_FILE")
        retval = self._ctx.hosts(filename)
        if retval != 0:
            location = filename or "system default hosts file"
            raise UnboundException(location, retval)

    @dns_enabled
    def _callback(self, req: DNSRequest, err: int, result):
        if err != 0 or (result.qtype == unbound.RR_TYPE_AAAA
                        and not result.havedata):
            req.state = ReqState.ERROR
            vlog.warn(f"{req.name}: failed to resolve")
            return
        if result.qtype == unbound.RR_TYPE_A and not result.havedata:
            self._resolve_async(req, unbound.RR_TYPE_AAAA)
            return
        try:
            ip_str = next(iter(result.data.as_raw_data()))
            ip = ipaddress.ip_address(ip_str)  # test if IP is valid
            # NOTE (twilson) For some reason, accessing result data outside of
            # _callback causes a segfault. So just grab and store what we need.
            req.result = str(ip)
            req.ttl = result.ttl
            req.state = ReqState.GOOD_UNUSED
            req.time = time.time()
        except (ValueError, StopIteration):
            req.state = ReqState.ERROR
            vlog.err(f"{req.name}: failed to resolve")

    @dns_enabled
    def _resolve_sync(self, name: str) -> typing.Optional[str]:
        for qtype in (unbound.RR_TYPE_A, unbound.RR_TYPE_AAAA):
            err, result = self._ctx.resolve(name, qtype)
            if err != 0:
                return None
            if not result.havedata:
                continue
            try:
                ip = ipaddress.ip_address(
                    next(iter(result.data.as_raw_data())))
            except (ValueError, StopIteration):
                return None
            return str(ip)

        return None

    @dns_enabled
    def _resolve_async(self, req: DNSRequest, qtype) -> None:
        err, _ = self._ctx.resolve_async(req.name, req, self._callback,
                                         qtype)
        if err != 0:
            req.state = ReqState.ERROR
            return None

        req.state = ReqState.PENDING
        return None

    @dns_enabled
    def resolve(self, name: str) -> typing.Optional[str]:
        """Resolve a host name to an IP address

        If the resolver is set to handle requests asynchronously, resolve()
        should be recalled until it returns a non-None result. Errors will be
        logged.

        :param name: The host name to resolve
        :returns: The IP address or None on error or not (yet) found
        """
        if not self._is_daemon:
            return self._resolve_sync(name)
        retval = self._ctx.process()
        if retval != 0:
            vlog.err(f"dns-resolve error: {unbound.ub_strerror(retval)}")
            return None
        req = self._requests[name]  # Creates a DNSRequest if not found
        if req.is_valid:
            return req.result
        elif req.state != ReqState.PENDING:
            self._resolve_async(req, unbound.RR_TYPE_A)
        return None


_global_resolver: typing.Optional[DNSResolver] = None


def init(is_daemon: bool = False) -> DNSResolver:
    """Initialize a global DNSResolver

    See DNSResolver.__init__ for more details
    """
    global _global_resolver
    _global_resolver = DNSResolver(is_daemon)
    return _global_resolver


def resolve(name: str) -> typing.Optional[str]:
    """Resolve a host name to an IP address

    If a DNSResolver instance has not been instantiated, or if it has been
    created with is_daemon=False, resolve() will synchronously resolve the
    hostname. If DNSResolver has been initialized with is_daemon=True, it
    will instead resolve asynchornously and resolve() will return None until
    the hostname has been resolved.

    :param name: The host name to resolve
    :returns: The IP address or None on error or not (yet) found
    """
    if _global_resolver is None:
        init()

    # mypy doesn't understand that init() sets _global_resolver, so ignore type
    return _global_resolver.resolve(name)  # type: ignore


def destroy():
    """Destroy the global DNSResolver

    This destroys the global DNSResolver instance and any outstanding
    asynchronouse requests.
    """
    global _global_resolver
    del _global_resolver
    _global_resolver = None  # noqa: F841



# Copyright (c) 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import click
import os

from ovs.flow.filter import OFFilter
from ovs.dirs import PKGDATADIR

_default_config_file = "ovs-flowviz.conf"
_default_config_path = next(
    (
        p
        for p in [
            os.path.join(
                os.getenv("HOME"), ".config", "ovs", _default_config_file
            ),
            os.path.join(PKGDATADIR, _default_config_file),
            os.path.abspath(
                os.path.join(os.path.dirname(__file__), _default_config_file)
            ),
        ]
        if os.path.exists(p)
    ),
    "",
)


class Options(dict):
    """Options dictionary"""


def validate_input(ctx, param, value):
    """Validate the "-i" option"""
    result = list()
    for input_str in value:
        parts = input_str.strip().split(",")
        if len(parts) == 2:
            file_parts = tuple(parts)
        elif len(parts) == 1:
            file_parts = tuple(["Filename: " + parts[0], parts[0]])
        else:
            raise click.BadParameter(
                "input filename should have the following format: "
                "[alias,]FILENAME"
            )

        if not os.path.isfile(file_parts[1]):
            raise click.BadParameter(
                "input filename %s does not exist" % file_parts[1]
            )
        result.append(file_parts)
    return result


@click.group(
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option(
    "-c",
    "--config",
    help="Use config file",
    type=click.Path(),
    default=_default_config_path,
    show_default=True,
)
@click.option(
    "--style",
    help="Select style (defined in config file)",
    default=None,
    show_default=True,
)
@click.option(
    "-i",
    "--input",
    "filename",
    help="Read flows from specified filepath. If not provided, flows will be"
    " read from stdin. This option can be specified multiple times."
    " Format [alias,]FILENAME. Where alias is a name that shall be used to"
    " refer to this FILENAME",
    multiple=True,
    type=click.Path(),
    callback=validate_input,
)
@click.option(
    "-f",
    "--filter",
    help="Filter flows that match the filter expression."
    "Run 'ovs-flowviz filter' for a detailed description of the filtering "
    "syntax",
    type=str,
    show_default=False,
)
@click.option(
    "-l",
    "--highlight",
    help="Highlight flows that match the filter expression."
    "Run 'ovs-flowviz filter' for a detailed description of the filtering "
    "syntax",
    type=str,
    show_default=False,
)
@click.pass_context
def maincli(ctx, config, style, filename, filter, highlight):
    """
    OpenvSwitch flow visualization utility.

    It reads openflow and datapath flows
    (such as the output of ovs-ofctl dump-flows or ovs-appctl dpctl/dump-flows)
    and prints them in different formats.
    """
    ctx.obj = Options()
    ctx.obj["filename"] = filename or None
    if filter:
        try:
            ctx.obj["filter"] = OFFilter(filter)
        except Exception as e:
            raise click.BadParameter("Wrong filter syntax: {}".format(e))

    if highlight:
        try:
            ctx.obj["highlight"] = OFFilter(highlight)
        except Exception as e:
            raise click.BadParameter("Wrong filter syntax: {}".format(e))

    config_file = config or _default_config_path
    parser = configparser.ConfigParser()
    parser.read(config_file)

    ctx.obj["config"] = parser
    ctx.obj["style"] = style


@maincli.command(hidden=True)
@click.pass_context
def filter(ctx):
    """
    \b
    Filter Syntax
    *************

     [! | not ] {key}[[.subkey]...] [OPERATOR] {value})] [LOGICAL OPERATOR] ...

    \b
    Comparison operators are:
        =   equality
        <   less than
        >   more than
        ~=  masking (valid for IP and Ethernet fields)

    \b
    Logical operators are:
        !{expr}:  NOT
        {expr} && {expr}: AND
        {expr} || {expr}: OR

    \b
    Matches and flow metadata:
        To compare against a match or info field, use the field directly, e.g:
            priority=100
            n_bytes>10
        Use simple keywords for flags:
            tcp and ip_src=192.168.1.1
    \b
    Actions:
        Actions values might be dictionaries, use subkeys to access individual
        values, e.g:
            output.port=3
        Use simple keywords for flags
            drop

    \b
    Examples of valid filters.
        nw_addr~=192.168.1.1 && (tcp_dst=80 || tcp_dst=443)
        arp=true && !arp_tsa=192.168.1.1
        n_bytes>0 && drop=true"""
    click.echo(ctx.command.get_help(ctx))


def main():
    """
    Main Function
    """
    maincli()




# This file is auto-generated. Do not edit!

from ovs.flow import decoders

field_decoders = {
    'dp_hash': decoders.Mask32,
    'recirc_id': decoders.decode_int,
    'packet_type': decoders.decode_default,
    'conj_id': decoders.decode_int,
    'tun_id': decoders.Mask64,
    'tunnel_id': decoders.Mask64,
    'tun_src': decoders.IPMask,
    'tun_dst': decoders.IPMask,
    'tun_ipv6_src': decoders.IPMask,
    'tun_ipv6_dst': decoders.IPMask,
    'tun_flags': decoders.decode_default,
    'tun_ttl': decoders.decode_int,
    'tun_tos': decoders.decode_int,
    'tun_gbp_id': decoders.Mask16,
    'tun_gbp_flags': decoders.Mask8,
    'tun_erspan_idx': decoders.decode_mask(20),
    'tun_erspan_ver': decoders.decode_mask(4),
    'tun_erspan_dir': decoders.decode_mask(1),
    'tun_erspan_hwid': decoders.decode_mask(6),
    'tun_gtpu_flags': decoders.Mask8,
    'tun_gtpu_msgtype': decoders.Mask8,
    'tun_metadata0': decoders.Mask992,
    'tun_metadata1': decoders.Mask992,
    'tun_metadata2': decoders.Mask992,
    'tun_metadata3': decoders.Mask992,
    'tun_metadata4': decoders.Mask992,
    'tun_metadata5': decoders.Mask992,
    'tun_metadata6': decoders.Mask992,
    'tun_metadata7': decoders.Mask992,
    'tun_metadata8': decoders.Mask992,
    'tun_metadata9': decoders.Mask992,
    'tun_metadata10': decoders.Mask992,
    'tun_metadata11': decoders.Mask992,
    'tun_metadata12': decoders.Mask992,
    'tun_metadata13': decoders.Mask992,
    'tun_metadata14': decoders.Mask992,
    'tun_metadata15': decoders.Mask992,
    'tun_metadata16': decoders.Mask992,
    'tun_metadata17': decoders.Mask992,
    'tun_metadata18': decoders.Mask992,
    'tun_metadata19': decoders.Mask992,
    'tun_metadata20': decoders.Mask992,
    'tun_metadata21': decoders.Mask992,
    'tun_metadata22': decoders.Mask992,
    'tun_metadata23': decoders.Mask992,
    'tun_metadata24': decoders.Mask992,
    'tun_metadata25': decoders.Mask992,
    'tun_metadata26': decoders.Mask992,
    'tun_metadata27': decoders.Mask992,
    'tun_metadata28': decoders.Mask992,
    'tun_metadata29': decoders.Mask992,
    'tun_metadata30': decoders.Mask992,
    'tun_metadata31': decoders.Mask992,
    'tun_metadata32': decoders.Mask992,
    'tun_metadata33': decoders.Mask992,
    'tun_metadata34': decoders.Mask992,
    'tun_metadata35': decoders.Mask992,
    'tun_metadata36': decoders.Mask992,
    'tun_metadata37': decoders.Mask992,
    'tun_metadata38': decoders.Mask992,
    'tun_metadata39': decoders.Mask992,
    'tun_metadata40': decoders.Mask992,
    'tun_metadata41': decoders.Mask992,
    'tun_metadata42': decoders.Mask992,
    'tun_metadata43': decoders.Mask992,
    'tun_metadata44': decoders.Mask992,
    'tun_metadata45': decoders.Mask992,
    'tun_metadata46': decoders.Mask992,
    'tun_metadata47': decoders.Mask992,
    'tun_metadata48': decoders.Mask992,
    'tun_metadata49': decoders.Mask992,
    'tun_metadata50': decoders.Mask992,
    'tun_metadata51': decoders.Mask992,
    'tun_metadata52': decoders.Mask992,
    'tun_metadata53': decoders.Mask992,
    'tun_metadata54': decoders.Mask992,
    'tun_metadata55': decoders.Mask992,
    'tun_metadata56': decoders.Mask992,
    'tun_metadata57': decoders.Mask992,
    'tun_metadata58': decoders.Mask992,
    'tun_metadata59': decoders.Mask992,
    'tun_metadata60': decoders.Mask992,
    'tun_metadata61': decoders.Mask992,
    'tun_metadata62': decoders.Mask992,
    'tun_metadata63': decoders.Mask992,
    'metadata': decoders.Mask64,
    'in_port': decoders.decode_default,
    'in_port_oxm': decoders.decode_default,
    'actset_output': decoders.decode_default,
    'skb_priority': decoders.decode_int,
    'pkt_mark': decoders.Mask32,
    'ct_state': decoders.decode_default,
    'ct_zone': decoders.decode_int,
    'ct_mark': decoders.Mask32,
    'ct_label': decoders.Mask128,
    'ct_nw_proto': decoders.decode_int,
    'ct_nw_src': decoders.IPMask,
    'ct_nw_dst': decoders.IPMask,
    'ct_ipv6_src': decoders.IPMask,
    'ct_ipv6_dst': decoders.IPMask,
    'ct_tp_src': decoders.Mask16,
    'ct_tp_dst': decoders.Mask16,
    'reg0': decoders.Mask32,
    'reg1': decoders.Mask32,
    'reg2': decoders.Mask32,
    'reg3': decoders.Mask32,
    'reg4': decoders.Mask32,
    'reg5': decoders.Mask32,
    'reg6': decoders.Mask32,
    'reg7': decoders.Mask32,
    'reg8': decoders.Mask32,
    'reg9': decoders.Mask32,
    'reg10': decoders.Mask32,
    'reg11': decoders.Mask32,
    'reg12': decoders.Mask32,
    'reg13': decoders.Mask32,
    'reg14': decoders.Mask32,
    'reg15': decoders.Mask32,
    'reg16': decoders.Mask32,
    'reg17': decoders.Mask32,
    'reg18': decoders.Mask32,
    'reg19': decoders.Mask32,
    'reg20': decoders.Mask32,
    'reg21': decoders.Mask32,
    'reg22': decoders.Mask32,
    'reg23': decoders.Mask32,
    'reg24': decoders.Mask32,
    'reg25': decoders.Mask32,
    'reg26': decoders.Mask32,
    'reg27': decoders.Mask32,
    'reg28': decoders.Mask32,
    'reg29': decoders.Mask32,
    'reg30': decoders.Mask32,
    'reg31': decoders.Mask32,
    'xreg0': decoders.Mask64,
    'xreg1': decoders.Mask64,
    'xreg2': decoders.Mask64,
    'xreg3': decoders.Mask64,
    'xreg4': decoders.Mask64,
    'xreg5': decoders.Mask64,
    'xreg6': decoders.Mask64,
    'xreg7': decoders.Mask64,
    'xreg8': decoders.Mask64,
    'xreg9': decoders.Mask64,
    'xreg10': decoders.Mask64,
    'xreg11': decoders.Mask64,
    'xreg12': decoders.Mask64,
    'xreg13': decoders.Mask64,
    'xreg14': decoders.Mask64,
    'xreg15': decoders.Mask64,
    'xxreg0': decoders.Mask128,
    'xxreg1': decoders.Mask128,
    'xxreg2': decoders.Mask128,
    'xxreg3': decoders.Mask128,
    'xxreg4': decoders.Mask128,
    'xxreg5': decoders.Mask128,
    'xxreg6': decoders.Mask128,
    'xxreg7': decoders.Mask128,
    'eth_src': decoders.EthMask,
    'dl_src': decoders.EthMask,
    'eth_dst': decoders.EthMask,
    'dl_dst': decoders.EthMask,
    'eth_type': decoders.decode_int,
    'dl_type': decoders.decode_int,
    'vlan_tci': decoders.Mask16,
    'dl_vlan': decoders.decode_int,
    'vlan_vid': decoders.decode_mask(12),
    'dl_vlan_pcp': decoders.decode_int,
    'vlan_pcp': decoders.decode_int,
    'mpls_label': decoders.decode_int,
    'mpls_tc': decoders.decode_int,
    'mpls_bos': decoders.decode_int,
    'mpls_ttl': decoders.decode_int,
    'ip_src': decoders.IPMask,
    'nw_src': decoders.IPMask,
    'ip_dst': decoders.IPMask,
    'nw_dst': decoders.IPMask,
    'ipv6_src': decoders.IPMask,
    'ipv6_dst': decoders.IPMask,
    'ipv6_label': decoders.decode_mask(20),
    'nw_proto': decoders.decode_int,
    'ip_proto': decoders.decode_int,
    'nw_tos': decoders.decode_int,
    'ip_dscp': decoders.decode_int,
    'nw_ecn': decoders.decode_int,
    'ip_ecn': decoders.decode_int,
    'nw_ttl': decoders.decode_int,
    'ip_frag': decoders.decode_default,
    'nw_frag': decoders.decode_default,
    'arp_op': decoders.decode_int,
    'arp_spa': decoders.IPMask,
    'arp_tpa': decoders.IPMask,
    'arp_sha': decoders.EthMask,
    'arp_tha': decoders.EthMask,
    'tcp_src': decoders.Mask16,
    'tp_src': decoders.Mask16,
    'tcp_dst': decoders.Mask16,
    'tp_dst': decoders.Mask16,
    'tcp_flags': decoders.decode_default,
    'udp_src': decoders.Mask16,
    'udp_dst': decoders.Mask16,
    'sctp_src': decoders.Mask16,
    'sctp_dst': decoders.Mask16,
    'icmp_type': decoders.decode_int,
    'icmp_code': decoders.decode_int,
    'icmpv6_type': decoders.decode_int,
    'icmpv6_code': decoders.decode_int,
    'nd_target': decoders.IPMask,
    'nd_sll': decoders.EthMask,
    'nd_tll': decoders.EthMask,
    'nd_reserved': decoders.decode_int,
    'nd_options_type': decoders.decode_int,
    'nsh_flags': decoders.Mask8,
    'nsh_mdtype': decoders.decode_int,
    'nsh_np': decoders.decode_int,
    'nsh_spi': decoders.decode_int,
    'nsp': decoders.decode_int,
    'nsh_si': decoders.decode_int,
    'nsi': decoders.decode_int,
    'nsh_c1': decoders.Mask32,
    'nshc1': decoders.Mask32,
    'nsh_c2': decoders.Mask32,
    'nshc2': decoders.Mask32,
    'nsh_c3': decoders.Mask32,
    'nshc3': decoders.Mask32,
    'nsh_c4': decoders.Mask32,
    'nshc4': decoders.Mask32,
    'nsh_ttl': decoders.decode_int,
}

field_aliases = {
    'NXOXM_ET_DP_HASH': 'dp_hash',
    'NXM_NX_DP_HASH': 'dp_hash',
    'NXM_NX_RECIRC_ID': 'recirc_id',
    'OXM_OF_PACKET_TYPE': 'packet_type',
    'NXM_NX_CONJ_ID': 'conj_id',
    'OXM_OF_TUNNEL_ID': 'tun_id',
    'NXM_NX_TUN_ID': 'tun_id',
    'NXM_NX_TUN_IPV4_SRC': 'tun_src',
    'NXM_NX_TUN_IPV4_DST': 'tun_dst',
    'NXM_NX_TUN_IPV6_SRC': 'tun_ipv6_src',
    'NXM_NX_TUN_IPV6_DST': 'tun_ipv6_dst',
    'NXM_NX_TUN_FLAGS': 'tun_flags',
    'NXM_NX_TUN_GBP_ID': 'tun_gbp_id',
    'NXM_NX_TUN_GBP_FLAGS': 'tun_gbp_flags',
    'NXOXM_ET_ERSPAN_IDX': 'tun_erspan_idx',
    'NXOXM_ET_ERSPAN_VER': 'tun_erspan_ver',
    'NXOXM_ET_ERSPAN_DIR': 'tun_erspan_dir',
    'NXOXM_ET_ERSPAN_HWID': 'tun_erspan_hwid',
    'NXOXM_ET_GTPU_FLAGS': 'tun_gtpu_flags',
    'NXOXM_ET_GTPU_MSGTYPE': 'tun_gtpu_msgtype',
    'NXM_NX_TUN_METADATA0': 'tun_metadata0',
    'NXM_NX_TUN_METADATA1': 'tun_metadata1',
    'NXM_NX_TUN_METADATA2': 'tun_metadata2',
    'NXM_NX_TUN_METADATA3': 'tun_metadata3',
    'NXM_NX_TUN_METADATA4': 'tun_metadata4',
    'NXM_NX_TUN_METADATA5': 'tun_metadata5',
    'NXM_NX_TUN_METADATA6': 'tun_metadata6',
    'NXM_NX_TUN_METADATA7': 'tun_metadata7',
    'NXM_NX_TUN_METADATA8': 'tun_metadata8',
    'NXM_NX_TUN_METADATA9': 'tun_metadata9',
    'NXM_NX_TUN_METADATA10': 'tun_metadata10',
    'NXM_NX_TUN_METADATA11': 'tun_metadata11',
    'NXM_NX_TUN_METADATA12': 'tun_metadata12',
    'NXM_NX_TUN_METADATA13': 'tun_metadata13',
    'NXM_NX_TUN_METADATA14': 'tun_metadata14',
    'NXM_NX_TUN_METADATA15': 'tun_metadata15',
    'NXM_NX_TUN_METADATA16': 'tun_metadata16',
    'NXM_NX_TUN_METADATA17': 'tun_metadata17',
    'NXM_NX_TUN_METADATA18': 'tun_metadata18',
    'NXM_NX_TUN_METADATA19': 'tun_metadata19',
    'NXM_NX_TUN_METADATA20': 'tun_metadata20',
    'NXM_NX_TUN_METADATA21': 'tun_metadata21',
    'NXM_NX_TUN_METADATA22': 'tun_metadata22',
    'NXM_NX_TUN_METADATA23': 'tun_metadata23',
    'NXM_NX_TUN_METADATA24': 'tun_metadata24',
    'NXM_NX_TUN_METADATA25': 'tun_metadata25',
    'NXM_NX_TUN_METADATA26': 'tun_metadata26',
    'NXM_NX_TUN_METADATA27': 'tun_metadata27',
    'NXM_NX_TUN_METADATA28': 'tun_metadata28',
    'NXM_NX_TUN_METADATA29': 'tun_metadata29',
    'NXM_NX_TUN_METADATA30': 'tun_metadata30',
    'NXM_NX_TUN_METADATA31': 'tun_metadata31',
    'NXM_NX_TUN_METADATA32': 'tun_metadata32',
    'NXM_NX_TUN_METADATA33': 'tun_metadata33',
    'NXM_NX_TUN_METADATA34': 'tun_metadata34',
    'NXM_NX_TUN_METADATA35': 'tun_metadata35',
    'NXM_NX_TUN_METADATA36': 'tun_metadata36',
    'NXM_NX_TUN_METADATA37': 'tun_metadata37',
    'NXM_NX_TUN_METADATA38': 'tun_metadata38',
    'NXM_NX_TUN_METADATA39': 'tun_metadata39',
    'NXM_NX_TUN_METADATA40': 'tun_metadata40',
    'NXM_NX_TUN_METADATA41': 'tun_metadata41',
    'NXM_NX_TUN_METADATA42': 'tun_metadata42',
    'NXM_NX_TUN_METADATA43': 'tun_metadata43',
    'NXM_NX_TUN_METADATA44': 'tun_metadata44',
    'NXM_NX_TUN_METADATA45': 'tun_metadata45',
    'NXM_NX_TUN_METADATA46': 'tun_metadata46',
    'NXM_NX_TUN_METADATA47': 'tun_metadata47',
    'NXM_NX_TUN_METADATA48': 'tun_metadata48',
    'NXM_NX_TUN_METADATA49': 'tun_metadata49',
    'NXM_NX_TUN_METADATA50': 'tun_metadata50',
    'NXM_NX_TUN_METADATA51': 'tun_metadata51',
    'NXM_NX_TUN_METADATA52': 'tun_metadata52',
    'NXM_NX_TUN_METADATA53': 'tun_metadata53',
    'NXM_NX_TUN_METADATA54': 'tun_metadata54',
    'NXM_NX_TUN_METADATA55': 'tun_metadata55',
    'NXM_NX_TUN_METADATA56': 'tun_metadata56',
    'NXM_NX_TUN_METADATA57': 'tun_metadata57',
    'NXM_NX_TUN_METADATA58': 'tun_metadata58',
    'NXM_NX_TUN_METADATA59': 'tun_metadata59',
    'NXM_NX_TUN_METADATA60': 'tun_metadata60',
    'NXM_NX_TUN_METADATA61': 'tun_metadata61',
    'NXM_NX_TUN_METADATA62': 'tun_metadata62',
    'NXM_NX_TUN_METADATA63': 'tun_metadata63',
    'OXM_OF_METADATA': 'metadata',
    'NXM_OF_IN_PORT': 'in_port',
    'OXM_OF_IN_PORT': 'in_port_oxm',
    'ONFOXM_ET_ACTSET_OUTPUT': 'actset_output',
    'OXM_OF_ACTSET_OUTPUT': 'actset_output',
    'NXM_NX_PKT_MARK': 'pkt_mark',
    'NXM_NX_CT_STATE': 'ct_state',
    'NXM_NX_CT_ZONE': 'ct_zone',
    'NXM_NX_CT_MARK': 'ct_mark',
    'NXM_NX_CT_LABEL': 'ct_label',
    'NXM_NX_CT_NW_PROTO': 'ct_nw_proto',
    'NXM_NX_CT_NW_SRC': 'ct_nw_src',
    'NXM_NX_CT_NW_DST': 'ct_nw_dst',
    'NXM_NX_CT_IPV6_SRC': 'ct_ipv6_src',
    'NXM_NX_CT_IPV6_DST': 'ct_ipv6_dst',
    'NXM_NX_CT_TP_SRC': 'ct_tp_src',
    'NXM_NX_CT_TP_DST': 'ct_tp_dst',
    'NXM_NX_REG0': 'reg0',
    'NXM_NX_REG1': 'reg1',
    'NXM_NX_REG2': 'reg2',
    'NXM_NX_REG3': 'reg3',
    'NXM_NX_REG4': 'reg4',
    'NXM_NX_REG5': 'reg5',
    'NXM_NX_REG6': 'reg6',
    'NXM_NX_REG7': 'reg7',
    'NXM_NX_REG8': 'reg8',
    'NXM_NX_REG9': 'reg9',
    'NXM_NX_REG10': 'reg10',
    'NXM_NX_REG11': 'reg11',
    'NXM_NX_REG12': 'reg12',
    'NXM_NX_REG13': 'reg13',
    'NXM_NX_REG14': 'reg14',
    'NXM_NX_REG15': 'reg15',
    'NXOXM_ET_REG16': 'reg16',
    'NXOXM_ET_REG17': 'reg17',
    'NXOXM_ET_REG18': 'reg18',
    'NXOXM_ET_REG19': 'reg19',
    'NXOXM_ET_REG20': 'reg20',
    'NXOXM_ET_REG21': 'reg21',
    'NXOXM_ET_REG22': 'reg22',
    'NXOXM_ET_REG23': 'reg23',
    'NXOXM_ET_REG24': 'reg24',
    'NXOXM_ET_REG25': 'reg25',
    'NXOXM_ET_REG26': 'reg26',
    'NXOXM_ET_REG27': 'reg27',
    'NXOXM_ET_REG28': 'reg28',
    'NXOXM_ET_REG29': 'reg29',
    'NXOXM_ET_REG30': 'reg30',
    'NXOXM_ET_REG31': 'reg31',
    'OXM_OF_PKT_REG0': 'xreg0',
    'OXM_OF_PKT_REG1': 'xreg1',
    'OXM_OF_PKT_REG2': 'xreg2',
    'OXM_OF_PKT_REG3': 'xreg3',
    'OXM_OF_PKT_REG4': 'xreg4',
    'OXM_OF_PKT_REG5': 'xreg5',
    'OXM_OF_PKT_REG6': 'xreg6',
    'OXM_OF_PKT_REG7': 'xreg7',
    'OXM_OF_PKT_REG8': 'xreg8',
    'OXM_OF_PKT_REG9': 'xreg9',
    'OXM_OF_PKT_REG10': 'xreg10',
    'OXM_OF_PKT_REG11': 'xreg11',
    'OXM_OF_PKT_REG12': 'xreg12',
    'OXM_OF_PKT_REG13': 'xreg13',
    'OXM_OF_PKT_REG14': 'xreg14',
    'OXM_OF_PKT_REG15': 'xreg15',
    'NXM_NX_XXREG0': 'xxreg0',
    'NXM_NX_XXREG1': 'xxreg1',
    'NXM_NX_XXREG2': 'xxreg2',
    'NXM_NX_XXREG3': 'xxreg3',
    'NXOXM_ET_XXREG4': 'xxreg4',
    'NXOXM_ET_XXREG5': 'xxreg5',
    'NXOXM_ET_XXREG6': 'xxreg6',
    'NXOXM_ET_XXREG7': 'xxreg7',
    'OXM_OF_ETH_SRC': 'eth_src',
    'NXM_OF_ETH_SRC': 'eth_src',
    'OXM_OF_ETH_DST': 'eth_dst',
    'NXM_OF_ETH_DST': 'eth_dst',
    'OXM_OF_ETH_TYPE': 'eth_type',
    'NXM_OF_ETH_TYPE': 'eth_type',
    'NXM_OF_VLAN_TCI': 'vlan_tci',
    'OXM_OF_VLAN_VID': 'vlan_vid',
    'OXM_OF_VLAN_PCP': 'vlan_pcp',
    'OXM_OF_MPLS_LABEL': 'mpls_label',
    'OXM_OF_MPLS_TC': 'mpls_tc',
    'OXM_OF_MPLS_BOS': 'mpls_bos',
    'NXM_NX_MPLS_TTL': 'mpls_ttl',
    'OXM_OF_IPV4_SRC': 'ip_src',
    'NXM_OF_IP_SRC': 'ip_src',
    'OXM_OF_IPV4_DST': 'ip_dst',
    'NXM_OF_IP_DST': 'ip_dst',
    'OXM_OF_IPV6_SRC': 'ipv6_src',
    'NXM_NX_IPV6_SRC': 'ipv6_src',
    'OXM_OF_IPV6_DST': 'ipv6_dst',
    'NXM_NX_IPV6_DST': 'ipv6_dst',
    'OXM_OF_IPV6_FLABEL': 'ipv6_label',
    'NXM_NX_IPV6_LABEL': 'ipv6_label',
    'OXM_OF_IP_PROTO': 'nw_proto',
    'NXM_OF_IP_PROTO': 'nw_proto',
    'NXM_OF_IP_TOS': 'nw_tos',
    'OXM_OF_IP_DSCP': 'ip_dscp',
    'OXM_OF_IP_ECN': 'nw_ecn',
    'NXM_NX_IP_ECN': 'nw_ecn',
    'NXM_NX_IP_TTL': 'nw_ttl',
    'NXM_NX_IP_FRAG': 'ip_frag',
    'OXM_OF_ARP_OP': 'arp_op',
    'NXM_OF_ARP_OP': 'arp_op',
    'OXM_OF_ARP_SPA': 'arp_spa',
    'NXM_OF_ARP_SPA': 'arp_spa',
    'OXM_OF_ARP_TPA': 'arp_tpa',
    'NXM_OF_ARP_TPA': 'arp_tpa',
    'OXM_OF_ARP_SHA': 'arp_sha',
    'NXM_NX_ARP_SHA': 'arp_sha',
    'OXM_OF_ARP_THA': 'arp_tha',
    'NXM_NX_ARP_THA': 'arp_tha',
    'OXM_OF_TCP_SRC': 'tcp_src',
    'NXM_OF_TCP_SRC': 'tcp_src',
    'OXM_OF_TCP_DST': 'tcp_dst',
    'NXM_OF_TCP_DST': 'tcp_dst',
    'ONFOXM_ET_TCP_FLAGS': 'tcp_flags',
    'OXM_OF_TCP_FLAGS': 'tcp_flags',
    'NXM_NX_TCP_FLAGS': 'tcp_flags',
    'OXM_OF_UDP_SRC': 'udp_src',
    'NXM_OF_UDP_SRC': 'udp_src',
    'OXM_OF_UDP_DST': 'udp_dst',
    'NXM_OF_UDP_DST': 'udp_dst',
    'OXM_OF_SCTP_SRC': 'sctp_src',
    'OXM_OF_SCTP_DST': 'sctp_dst',
    'OXM_OF_ICMPV4_TYPE': 'icmp_type',
    'NXM_OF_ICMP_TYPE': 'icmp_type',
    'OXM_OF_ICMPV4_CODE': 'icmp_code',
    'NXM_OF_ICMP_CODE': 'icmp_code',
    'OXM_OF_ICMPV6_TYPE': 'icmpv6_type',
    'NXM_NX_ICMPV6_TYPE': 'icmpv6_type',
    'OXM_OF_ICMPV6_CODE': 'icmpv6_code',
    'NXM_NX_ICMPV6_CODE': 'icmpv6_code',
    'OXM_OF_IPV6_ND_TARGET': 'nd_target',
    'NXM_NX_ND_TARGET': 'nd_target',
    'OXM_OF_IPV6_ND_SLL': 'nd_sll',
    'NXM_NX_ND_SLL': 'nd_sll',
    'OXM_OF_IPV6_ND_TLL': 'nd_tll',
    'NXM_NX_ND_TLL': 'nd_tll',
    'ERICOXM_OF_ICMPV6_ND_RESERVED': 'nd_reserved',
    'ERICOXM_OF_ICMPV6_ND_OPTIONS_TYPE': 'nd_options_type',
    'NXOXM_NSH_FLAGS': 'nsh_flags',
    'NXOXM_NSH_MDTYPE': 'nsh_mdtype',
    'NXOXM_NSH_NP': 'nsh_np',
    'NXOXM_NSH_SPI': 'nsh_spi',
    'NXOXM_NSH_SI': 'nsh_si',
    'NXOXM_NSH_C1': 'nsh_c1',
    'NXOXM_NSH_C2': 'nsh_c2',
    'NXOXM_NSH_C3': 'nsh_c3',
    'NXOXM_NSH_C4': 'nsh_c4',
    'NXOXM_NSH_TTL': 'nsh_ttl',
}
